/*
/*  ARadEngn - Main program to run the Airrad 8.0 computational engine
/*             in batch mode.
/*
/*  Author: Fred Wasmer
/*          Wasmer Consulting
/*
/*  Language: ANSI C
/*
/*  Translated: July, 1994
/*
/*  Purpose: ARadEngn generates a file of radiation intensity grids from which
/*           a variety of radiation fallout hazard indicators can be derived.
/*           This program performs the same task as the computational engine in
/*           the Airrad 8.0 model.
/*
/*  Airrad 8.0 is an IBM-PC DOS application used to calculate and display
/*  a variety of radiation fallout hazard indicators resulting from a user-
/*  specified nuclear burst scenario.  Airrad was written in the late 1980's
/*  using Turbo Pascal.
/*
/*  This program is a port to ANSI C of the Airrad 8.0 computation engine.
/*
/*  This program takes two command-line options: the name of a ASCII Airrad
/*  case description input file, and the name of a binary grid output file.
/*  If these are not specified, then by default, "airrad.cas" and "airrad.grd"
/*  are used.
/*
/*  The ASCII Airrad case description file contains a description of an Airrad
/*  scenario: burst characteristics (burst size, fission type, etc.), a wind
/*  profile (wind speed and direction as a function of height), grid information
/*  (the size, resolution, and location of the grid on which computations are
/*  performed), a an optional list of specific receptors locations.
/*
/*  The binary grid file contains a series of grids from which a variety of
/*  radiation hazard indicators (exposure rate, dosage, time of last fallout,
/*  etc.) can be computed.
/*
/*  This program reads in the Airrad case description file, performs the
/*  computations, then writes the binary grid file.
/*
/*  After the binary grid file has been created for a case, specific hazard
/*  indicators can be obtained by using the functions exported in the hazsubs.c
/*  file.
/*
/**/


#include "aradengn.h"


int main (int argc, char* argv[])

{
    char CaseFileName [128];
    char GridFileName [128];


    if (argc == 1)  /*  Use default files  */
    {
        strcpy (CaseFileName, "airrad.cas");
        strcpy (GridFileName, "airrad.grd");
    }
    else if (argc == 3)  /*  User has supplied file name on command line.  */
    {
        strcpy (CaseFileName, argv [1]);
        strcpy (GridFileName, argv [2]);
    }
    else
    {
        printf ("usage: aradengn <inflnm> <outflnm> \n");
        printf ("\n");
        printf ("       <inflnm> = Name of input file containing case for\n");
        printf ("                  which computations are to be performed.\n");
        printf ("\n");
        printf ("       <outflnm> = Name of binary output file that stores\n");
        printf ("                   results of computations.\n");

        FailExit ("Improper number of command line arguments");
    };


    #ifdef Debug
      printf ("\nRunning Airrad Computational Engine\n\n");
      printf ("Requested case file name = %s\n", CaseFileName);
      printf ("Requested grid file name = %s\n", GridFileName);
      printf ("\n\n");
    #endif


    PerformRoundingTests ();

    ReadCaseVariables (CaseFileName);

    PerformComputations ();

    WriteComputedGrids (GridFileName);


    SuccessExit ();

    return (0);

};  /*  int main  */








