

/*  The following is for setting up a debugging environment  */
/*  under Borland C++ 4.0.                                   */

#ifdef EasyWin

    #define  wait  getch ()

    #include <conio.h>

#else

    #ifdef __BORLANDC__

        #define ___DEFS_H
        #define _FAR
        #define _RTLENTRY
        #define _EXPFUNC
        #define _EXPFUNC16
        #define _EXPFUNC32
        #define _RTLENTRYF
        #define _USERENTRY

        int _RTLENTRY getch( void );

        #define wait getch ()

    #else

        #define wait

    #endif

#endif




/*  The following are the ANSI C standard libraries needed by the  */
/*  Airrad computational engine.                                   */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>




/* #define  Debug */

    /*  If Debug is defined, then a complete record of progress  */
    /*  by the model is output to standard out.                  */


#define  NumberComputedGrids  16

    /*  Number of samples of radiation pattern     */
    /*  computed during a fallout event.           */
    /*  This is the number of frame grids (see     */
    /*  the Airrad 8.0 Tech Report for definition  */
    /*  of frame grid.                             */


#define  MaxXGridSize  100
#define  MaxYGridSize  100

    /*  Size of the two-d arrays which contain the computed fallout grids.  */


#define  MaxReceptors  MaxXGridSize

    /*  The maximum number of specific receptors.  Since the    */
    /*  grid data for these points is stored along one side of  */
    /*  the grid, MaxReceptors must equal MaxXGridSize.         */


#define  MaxWindLayers  60

    /*  10 + The maximum number of wind measurement layers in the        */
    /*  atmospheric sounding.  The extra 10 array cells are work space.  */




#define  Pi        3.141592654
#define  HalfPi    (Pi/2.0)
#define  TwoPi     (Pi*2.0)
#define  DegToRad  (Pi/180.0)
#define  RadToDeg  (180.0/Pi)

#define  Min(A,B)  (((A)<(B)) ? (A) : (B))
#define  Max(A,B)  (((A)>(B)) ? (A) : (B))

    /*  Math constants and functions  */




#define  Round(X)      ( ceil ((X) - 0.5) )
#define  RoundUp(X)    ( ceil     (X)     )
#define  RoundDown(X)  ( floor    (X)     )
#define  RoundIn(X)    ( (X) < 0.0  ?  ceil  (X)  :  floor (X) )
#define  RoundOut(X)   ( (X) < 0.0  ?  floor (X)  :  ceil  (X) )

    /*  These five functions implement various methods of rounding
    /*  a floating point value to an integer.
    /*
    /*  Round (X) rounds to the closest integer.  Equivalent to Pascal's
    /*  Round function.
    /*
    /*  RoundUp (X) rounds to the smallest integer greater than or equal to X.
    /*
    /*  RoundDown (X) rounds to the largest integer less than or equal to X.
    /*
    /*  RoundIn (X) rounds to the closest integer towards zero: that is, to the
    /*  integer I that a) minimizes Abs (X-I), and b) meets the criteria
    /*  Abs (I) >= Abs (X).  Equivalent to Pascal's Trunc function.
    /*
    /*  RoundOut (X) rounds to the closest integer away from zero: that is, to
    /*  the integer I that a) minimizes Abs (X - I), and b) meets the criteria
    /*  Abs (I) <= Abs (X).
    /*
    /*                                         Value of X
    /*                   -1.2   -1    -.7   -.2     0    .2    .7     1    1.2
    /*                  -------------------------------------------------------
    /*    Round     (X)   -1    -1     -1     0     0     0     1     1     1
    /*    RoundUp   (X)   -1    -1      0     0     0     1     1     1     2
    /*    RoundDown (X)   -2    -1     -1    -1     0     0     0     1     1
    /*    RoundIn   (X)   -1    -1      0     0     0     0     0     1     1
    /*    RoundOut  (X)   -2    -1     -1    -1     0     1     1     1     2
    /*
    /**/




/*  Definition of data structures and variables to hold grid data  */

typedef float GridArray [MaxXGridSize] [MaxYGridSize + 1];

    /*  Note: We define an extra line in the Y direction  */
    /*        to hold specific receptor data.             */


typedef GridArray* GridArrayPtr;

typedef struct GridStruct
        {
            double Code;
            double Time;

            int NumXGridPoints,
                NumYGridPoints,
                NumReceptors;

            GridArray Grid;

            struct GridStruct* Next;

        } GridRecord;  /*  struct GridRecord  */

typedef GridRecord* GridPointer;

extern GridPointer GridListHead;

extern double GridRotationAngle;




/*  The following declerations are exported for use by the testing programs  */

#define  MetUndefined(X)  (fabs ((X)-999) < 0.001)

extern double ZGZ;

extern int NumberOfLayers;

typedef double WindArrayType [MaxWindLayers];

extern WindArrayType WindMeasurementHeight;
extern WindArrayType WindSpeed;
extern WindArrayType WindDirection;
extern WindArrayType Uwind;
extern WindArrayType Vwind;
extern WindArrayType WindMeasurementAltitude;




/*  The following functions are shared between various  */
/*  files of the Airrad ANSI-C Computational Engine.    */

typedef double MathDoubleFuncDoubleType (double);


void SuccessExit (void);

void FailExit (char *Message);

void PerformRoundingTests (void);

void Hunt (double XX [], int N, double X, int* Jlo);

void NewSpline (double X[], double Y[], int N, double YP1, double YPn,
                double Y2[]);

double Spline (double Xa[], double Ya[], double Y2a[], int N, double X);

double LinearInterpolate (double X[], double Y[], int N, double Xeval);

void HeapSort (double A[], int N, int SortOrderSwitch);

double Integrate (MathDoubleFuncDoubleType Fun, double A, double B,
                  double AbsErr, double RelErr);

GridPointer ZeroGrids (GridPointer GridListHead,
                       int NumberGrids, int NumXGridPoints, int NumYGridPoints,
                       int NumReceptors);

void ProcessWindData (void);

void ReadGrid (FILE* f, GridPointer GP);




/*  The following functions are intended to be exported out of the  */
/*  Airrad ANSI-C compuational engine, and called by an outside     */
/*  enclosing program.                                              */

void ReadCaseVariables (char *FileName);

void PerformComputations (void);

void WriteComputedGrids (char* FileName);

void ReadComputedGrids (char* FileName);

void UpdateRunProgressIndicator (int Stage, float PercentComplete);

void ComputeHazard (int HazardCode, double T1, double T2, double Height,
                    int RadUnits, GridArray Grid, double* GridAngle);




