
#include "aradengn.h"




GridPointer GridListHead = NULL;

double GridRotationAngle = 0.0;








void SuccessExit ()

/*  Called when a normal termination point has been reached
/*  without any errors.  Sends the word SUCCESS to standard
/*  out, which signals that no execution has completed and
/*  that no errors occured.
/*
/**/

{
    printf ("\n\nSUCCESS\n");

    exit (0);           

};  /*  void SuccessExit  */








void FailExit (char *Message)

/*  Called when an error occurs.  Sends the word FAIL
/*  to standard out, which signals that an error has
/*  occured.  Then sends the short error message pointed to
/*  by Message, enclosed in single quotes ''.
/*
/**/

{
    printf ("\n\nFAIL '%s'\n", Message);

    exit (1);

};  /*  void FailExit  */








void PerformRoundingTests ()

/*  Tests the five rounding macros Round, RoundUp, RoundDown,
/*  RoundIn, and RoundOut, to insure that they are properly
/*  implemented on the current platform.
/*
/**/

{
    int i;

    typedef struct
    {
        double X;
        int    ValueForRound;
        int    ValueForRoundUp;
        int    ValueForRoundDown;
        int    ValueForRoundOut;
        int    ValueForRoundIn;
    } TestStruct;


    TestStruct Data [] =
        {
          {-1.2,  -1, -1, -2, -2, -1},
          {-1.001,-1, -1, -2, -2, -1},
          {-1.0,  -1, -1, -1, -1, -1},
          {-0.999,-1,  0, -1, -1,  0},
          {-0.7,  -1,  0, -1, -1,  0},
          {-0.501,-1,  0, -1, -1,  0},
          {-0.499, 0,  0, -1, -1,  0},
          {-0.2,   0,  0, -1, -1,  0},
          {-0.001, 0,  0, -1, -1,  0},
          { 0.0,   0,  0,  0,  0,  0},
          { 0.001, 0,  1,  0,  1,  0},
          { 0.2,   0,  1,  0,  1,  0},
          { 0.499, 0,  1,  0,  1,  0},
          { 0.501, 1,  1,  0,  1,  0},
          { 0.7,   1,  1,  0,  1,  0},
          { 0.999, 1,  1,  0,  1,  0},
          { 1.0,   1,  1,  1,  1,  1},
          { 1.001, 1,  2,  1,  2,  1},
          { 1.2,   1,  2,  1,  2,  1},
          { 0.0, -99,-99,-99,-99,-99}
        };


    #ifdef Debug
      printf ("\nTesting Rounding functions\n");
    #endif


    for (i=0; Data [i].ValueForRound != -99; i++)
    {
        if (Round (Data [i].X) != Data [i].ValueForRound)
            FailExit ("Round macro improperly defined");

        if (RoundUp (Data [i].X) != Data [i].ValueForRoundUp)
            FailExit ("RoundUp macro improperly defined");

        if (RoundDown (Data [i].X) != Data [i].ValueForRoundDown)
            FailExit ("RoundDown macro improperly defined");

        if (RoundIn (Data [i].X) != Data [i].ValueForRoundIn)
            FailExit ("RoundIn macro improperly defined");

        if (RoundOut (Data [i].X) != Data [i].ValueForRoundOut)
            FailExit ("RoundOut macro improperly defined");

    };

};  /*  void PerformRoundingTests  */








GridPointer ZeroGrids (GridPointer GridListHead,
                       int NumberGrids, int NumXGridPoints, int NumYGridPoints,
                       int NumReceptors)

/*  Call with GridListHead, which points to either nil or a linked list of
/*  grids generated with a previous call to ZeroGrids.  Returns by
/*  GridListHead a linked list of NumberGrids grid records, each with the
/*  1..NumXGridPoints, 1..NumYGridPoints domain of their grid field set to
/*  0.0's.
/*
/*  Returns the new values for GridListHead.
/*
/**/

{
    int N, IX, IY, IReceptor;

    GridPointer GridP;


    #ifdef Debug
      printf ("Zeroing all grids\n");
    #endif


    if (GridListHead == NULL)
    {
        GridListHead       = malloc (sizeof (*GridListHead));
        GridListHead->Next = NULL;
    };

    GridP = GridListHead;


    for (N = 1; N <= NumberGrids; N++)
    {
        #ifdef Debug
          printf ("   - Zeroing grid %i\n", N);
        #endif

        for (IX = 0; IX < NumXGridPoints; IX++)
            for (IY = 0; IY < NumYGridPoints; IY++)
                GridP->Grid [IX][IY] = 0.0;

        for (IReceptor = 0; IReceptor < NumReceptors; IReceptor++)
           GridP->Grid [IReceptor][NumYGridPoints] = 0.0;

        GridP->NumXGridPoints = NumXGridPoints;
        GridP->NumYGridPoints = NumYGridPoints;
        GridP->NumReceptors   = NumReceptors;


        if (N < NumberGrids)
        {
            if (GridP->Next == NULL)
            {
                GridP->Next       = malloc (sizeof (*GridP));
                GridP->Next->Next = NULL;
             };

            GridP = GridP->Next;
        };
    };


    #ifdef Debug
      printf ("\n\n");
    #endif


    return (GridListHead);

};  /*  GridPointer ZeroGrids  */








void Hunt (double XX [], int N, double X, int* Jlo)

/*   Pass an Double array XX [N], and a desired target X.
/*   Returns Jlo such that XX [Jlo] <= X <= XX [Jlo + 1].
/*   If several calls are made to hunt with X's close to each other,
/*   do not change the value of Jlo between calls.  The Hunt algorithm
/*   employed is much faster than a bisection search when the value of
/*   Jlo passed to Hunt is close to the desired value.
/*
/**/

{
    int Jm, Jhi, Inc;

    int Ascnd, StopLoop;


    Ascnd = XX [N-1] > XX [0];

    if ((*Jlo <= -1) || (*Jlo > (N-1)))
    {
      *Jlo = -1;
      Jhi  =  N;
    }
    else
    {

        Inc      = 1;
        StopLoop = 0;

        if ((X >= XX [*Jlo]) == Ascnd)
            do
            {
                Jhi = *Jlo + Inc;

                if (Jhi > (N-1))
                {
                    Jhi      = N;
                    StopLoop = 1;
                }
                else
                    if ((X >= XX [Jhi])==Ascnd)
                    {
                        *Jlo = Jhi;
                         Inc = 2*Inc;
                    }
                    else
                        StopLoop = 1;

            } while (! StopLoop);

        else
        {

            Jhi = *Jlo;

            do
            {
                *Jlo = Jhi - Inc;

                if (*Jlo < 0)
                {
                    *Jlo      = -1;
                     StopLoop = 1;
                }
                else
                    if ((X < XX [*Jlo]) == Ascnd)
                    {
                        Jhi = *Jlo;
                        Inc = 2*Inc;
                    }
                    else
                        StopLoop = 1;

            } while (!StopLoop);

        };

    };


    while ((Jhi - *Jlo) != 1)
    {
        Jm = (Jhi + *Jlo) / 2;

        if ((X > XX [Jm]) == Ascnd)
            *Jlo = Jm;
        else
            Jhi = Jm;
    };

};  /*  void Hunt  */








void NewSpline (double X[], double Y[], int N, double YP1, double YPn,
                double Y2[])

/*   Compute and return in array Y2 the second derivatives used by function
/*   Spline to perform a spline interpolation of the function F, Yi := F (Xi).
/*
/*   X   = Array containing the N points of the independent variable.
/*         (X0 < X1 < ... < X(n-1))
/*   Y   = Array containing the N points of the dependent variable.
/*   N   = Number of points in X and Y.
/*   YP1,
/*   YPn = First derivatives of the function F at point 1 and N.  YP1 and/or
/*         YPn >= 1.0e30 is a signal to set the corresponding boundary
/*         conditions for a natural spline (second derivative = 0.0).
/*
/**/

{
    int i;

    double P, Qn, Sig, Un;

    double (*Up) [];


    if (N < 2) return;

    Up = malloc (sizeof (double) * N);


    if (YP1 > 0.99e30)
    {
        Y2     [0] = 0.0;
        (*Up)  [0] = 0.0;
    }
    else
    {
        Y2    [0] = -0.5;
        (*Up) [0] = (3.0/(X[1] - X[0]))*((Y[1] - Y[0])/(X[1] - X[0]) - YP1);
    };


    for (i = 1; i < (N-1); i++)
    {
        Sig       = (X[i] - X[i-1])/(X[i + 1] - X[i-1]);
        P         = Sig*Y2[i-1] + 2.0;
        Y2 [i]    = (Sig - 1.0)/P;
        (*Up) [i] = (Y[i+1] - Y[i  ]) / (X[i+1] - X[i  ]) -
                    (Y[i  ] - Y[i-1]) / (X[i  ] - X[i-1]);
        (*Up) [i] = (6.0*(*Up)[i]/(X[i+1] - X[i-1]) - Sig*(*Up)[i-1])/P;
    };


    if (YPn > 0.99e30)
    {
        Qn = 0.0;
        Un = 0.0;
    }
    else
    {
        Qn = 0.5;
        Un = (3.0/(X[N-1] - X[N-2]))*(YPn - (Y[N-1] - Y[N-2]) /
             (X[N-1] - X[N-2]));
    };

    Y2 [N-1] = (Un - Qn*(*Up)[N-2])/(Qn*Y2[N-2] + 1.0);

    for (i = (N-2); i >= 0; i--)
        Y2 [i] = Y2 [i]*Y2[i+1] + (*Up) [i];

    free (Up);

};  /*  void NewSpline  */








double Spline (double Xa[], double Ya[], double Y2a[], int N, double X)

/*   Returns the cubic-spline interpolation of F at point X.
/*   Xa, Ya, Y2a, and N correspond to the variables X, Y, Y2, and N in the
/*   routine NewSpline, which must be called once before this function.
/*
/*   Calls procedure Hunt.
/*
/**/

{
    static int Klo = 0;

    int Khi;

    double A, B, h;


    Hunt (Xa, N, X, &Klo);

    Khi = Klo + 1;

    h = Xa [Khi] - Xa [Klo];

    A = (Xa [Khi] - X) / h;
    B = (X - Xa [Klo]) / h;

    return (A*Ya [Klo] + B*Ya [Khi] + ((A*A*A - A)*Y2a [Klo] +
            (B*B*B - B)*Y2a [Khi])*(h*h)/6.0);

};  /*  double Spline  */








double LinearInterpolate (double X[], double Y[], int N, double Xeval)

/*   Returns the linear interpolation of the function F at point Xeval.
/*
/*   N = Number of points in X and Y.
/*
/*   X = Array containing the N increasing points of the independent variable.
/*       (X0 < X1 < ... < X(n-1))
/*
/*   Y = Array containing the N points of the dependent variable.
/*       Yi = F (Xi) for 0 <= i <= (N-1).
/*
/*   If Xeval < X0,     then Y0     is returned.
/*   If Xeval > X(n-1), then Y(n-1) is returned.
/*
/*   Calls procedure Hunt.
/*
/**/

{
    static int Klo = 0;

    int Khi;


    if (Xeval <= X [0])
        return (Y [0]);
    else
        if (Xeval >= X [N-1])
            return (Y [N-1]);
        else
        {
            Hunt (X, N, Xeval, &Klo);

            Khi = Klo + 1;

            return (Y [Klo] + (Y [Khi] - Y [Klo])*(Xeval - X [Klo]) /
                                      (X [Khi] - X [Klo]));
        };

};  /*  double LinearInterpolate  */








int (*SortOdrP) [] = NULL;




void HeapSort (double A[], int N, int SortOrderSwitch)

/*  Performs a heap sort on a double or integer array.
/*
/*  A is the array of N elements to be sorted.
/*
/*  SortOrderSwitch may be...
/*
/*     0: No sort order record is created.
/*
/*     1: An dynamic array of N word elements is created.  Into this
/*        array is stored a sort order record of the sorting process of A.
/*
/*     2: It is assumed that HeapSort has been called
/*        previously with SortOrderSwitch = 1.  The array A is sorted into the
/*        order defined by the previously created sort order record.
/*
/*     3: Like two, except the sort order array is destroyed after the sort,
/*        returning the memory to the heap.
/*
/**/

{
    double (*RTemp) [];

    int I, J, L, Ir, Io;

    double Rt;

    int Sorted;


    if (N <= 0) return;


    if (SortOrderSwitch == 1)
    {
        SortOdrP = malloc (sizeof (int) * N);

        for (I = 0; I < N; I++)
            (*SortOdrP) [I] = I;
    };


    if ((SortOrderSwitch == 0) || (SortOrderSwitch == 1))
    {
      L      = (N/2) + 1;
      Ir     = N;
      Sorted = 0;
    };


    if ((SortOrderSwitch == 2) || (SortOrderSwitch == 3))
    {
        RTemp = malloc (sizeof (double) * N);

        for (I = 0; I < N; I++)
            (*RTemp) [I] = A [(*SortOdrP) [I]];

        for (I = 0; I < N; I++)
            A [I] = (*RTemp) [I];

        free (RTemp);

        if (SortOrderSwitch == 3)
        {
            free (SortOdrP);
            SortOdrP = NULL;
        };

        return;

    };



    do
    {
        if (L > 1)
        {
            L--;

            Rt = A [L-1];

            if (SortOrderSwitch == 1)
                Io = (*SortOdrP) [L-1];
        }
        else
        {
            Rt = A [Ir-1];

            if (SortOrderSwitch == 1)
                Io = (*SortOdrP) [Ir-1];

            A [Ir-1] = A [0];

            if (SortOrderSwitch == 1)
                (*SortOdrP) [Ir-1] = (*SortOdrP) [0];

            Ir--;

            if (Ir <= 1)
            {
                A [0] = Rt;

               if (SortOrderSwitch == 1)
                   (*SortOdrP) [0] = Io;

               Sorted = 1;
            };
        };

        if (! Sorted)
        {

            I = L;
            J = 2*L;

            while (J <= Ir)
            {
                if (J < Ir)
                    if (A [J-1] < A [J])
                        J++;

                if (Rt < A [J-1])
                {
                    A [I-1] = A [J-1];

                    if (SortOrderSwitch == 1)
                        (*SortOdrP) [I-1] = (*SortOdrP) [J-1];

                    I = J;
                    J = 2*J;
                }
                else
                    J = Ir + 1;
            };

            A [I-1] = Rt;

            if (SortOrderSwitch == 1)
                (*SortOdrP) [I-1] = Io;

        };

    } while (! Sorted);

};  /*  void HeapSort */








double Integrate (MathDoubleFuncDoubleType Fun, double A, double B,
                  double AbsErr, double RelErr)

/*  Integrate the function Fun from A to B.  Fun should be a function which
/*  takes a single double precision argument and returns a double precision
/*  result.
/*
/*  The range of integration is from A to B. AbsErr and RelErr are
/*  the desired absolute and relative errors of the result.
/*
/*  The computed integral is returned functionally.
/*
/*  Example... Print the integral from -1.0 to 2.0 of X^2 + 1.
/*
/*
/*  Step One: Define the function to be integrated.
/*
/*  double UsersFunction (double X)
/*  {
/*      return (X*X + 1.0);
/*  };
/*
/*
/*  Step Two: Add call to integral into user's program.
/*
/*     {     /*  Main Code  */
/*
/*          printf ("The integral from -1.0 to 2.0 of X^2 + 1 = %lf",
/*                   Integrate (UsersFunction, -1.0, 2.0, 0.001, 0.001));
/*
/*     };      /*  Main Code  */
/*
/**/

{
    int i,
        j,
        Lev,
        Nim,
        NoFun,
        NoMax,
        NoFin,
        LevMin,
        LevMax,
        LevOut;

    double W0,
           W1,
           W2,
           W3,
           W4,
           Area,
           Temp,
           QNow,
           Step,
           Stone,
           Cor11,
           QPrev,
           QDiff,
           QLeft,
           EstErr,
           TolErr,
           ErrEst,
           Result;

    static double F [17], X [17];

    static double QRight [32];

    static double FSave [9][31];
    static double XSave [9][31];


    /**/
    /*
    /*  Stage 1 - General Initilization
    /*
    /*  Set Constants
    /*
    /**/

    LevMin = 1;
    LevMax = 30;
    LevOut = 6;
    NoMax  = 5000;

    NoFin = 1;

    for (i = 1; i <= (LevOut + 1); i++)
        NoFin = NoFin * 2;

    NoFin = NoMax - 8*(LevMax - LevOut + NoFin);

        /*  Trouble if NoFun reaches NoFin  */


    W0 =   3956.0/14175.0;
    W1 =  23552.0/14175.0;
    W2 =  -3712.0/14175.0;
    W3 =  41984.0/14175.0;
    W4 = -18160.0/14175.0;


    /*  Initialize Running Sums to Zero  */

    Result = 0.0;
    Cor11  = 0.0;
    ErrEst = 0.0;
    Area   = 0.0;
    NoFun  = 0;

    if (A == B)
        return (Result);


    /**/
    /*
    /*  Stage 2 - Initialization for First Interval
    /*
    /**/

    Lev    = 0;
    Nim    = 1;
    X [ 0] = A;
    X [16] = B;
    QPrev  = 0.0;
    F [0]  = Fun (X [0]);
    Stone  = (B-A)/16.0;

    X [ 8] = (X [ 0] + X [16])/2.0;
    X [ 4] = (X [ 0] + X [ 8])/2.0;
    X [12] = (X [ 8] + X [16])/2.0;
    X [ 2] = (X [ 0] + X [ 4])/2.0;
    X [ 6] = (X [ 4] + X [ 8])/2.0;
    X [10] = (X [ 8] + X [12])/2.0;
    X [14] = (X [12] + X [16])/2.0;

    for (i = 1; i <= 8; i++)
        F [i*2] = Fun (X [i*2]);

    NoFun = 9;


    L30: /**/
         /*
         /*  Stage 3 - Central Calculation
         /*
         /*  Requires QPrev, X0, X2, X4, ..., X16, F0, F2, F4, ..., F16.
         /*
         /*  Calculates X1, X3, ..., X15, F1, F3, ..., F15, QLeft, QRight, QNow, QDiff, Area.
         /*
         /**/

    X [1] = (X [0] + X [2])/2.0;

    F [1] = Fun (X [1]);

    for (i = 1; i <= 7; i++)
    {
        j = (i*2) + 1;

        X [j] = (X [j-1] + X [j + 1])/2.0;
        F [j] = Fun (X [j]);
    };

    NoFun = NoFun + 8;
    Step  = (X [16] - X [0])/16.0;

    QLeft = Step * (W0*(F [0] + F [8]) +
                    W1*(F [1] + F [7]) +
                    W2*(F [2] + F [6]) +
                    W3*(F [3] + F [5]) +
                    W4* F [4]);

    QRight [Lev + 1] = Step * (W0*(F [ 8] + F [16]) +
                               W1*(F [ 9] + F [15]) +
                               W2*(F [10] + F [14]) +
                               W3*(F [11] + F [13]) +
                               W4* F [12]);

    QNow  = QLeft + QRight [Lev + 1];
    QDiff = QNow - QPrev;
    Area  = Area + QDiff;


    /**/
    /*
    /*  Stage 4 - Interval Convergence Test
    /*
    /**/

    EstErr = fabs (QDiff)/1023.0;
    TolErr = Max (AbsErr, RelErr*fabs (Area)) * (Step/Stone);

    if (Lev     < LevMin)  goto L50;
    if (Lev    >= LevMax)  goto L70;
    if (NoFun   > NoFin)   goto L60;
    if (EstErr <= TolErr)  goto L70;


    L50: /**/
         /*
         /*  Stage 5 - No Convergence, Locate Next Interval
         /*
         /**/

    Nim = 2*Nim;

    Lev++;


    /*  Store Right Hand Elements for Future Use  */

    for (i = 1; i <= 8; i++)
    {
        FSave [i][Lev] = F [i + 8];
        XSave [i][Lev] = X [i + 8];
    };


    /*  Assemble Left Hand Elements for Immediate Use  */

    QPrev = QLeft;

    for (i = 1; i <= 8; i++)
    {
        j = -i;

        F [2*j + 18] = F [j + 9];
        X [2*j + 18] = X [j + 9];
    };

    goto L30;


    L60: /**/
         /*
         /*  Stage 6 - Trouble Section, Number of Function Values is About to Exceed Limit
         /*
         /**/

    NoFin  = 2*NoFin;
    LevMax = LevOut;


    L70: /**/
         /*
         /*  Stage 7 - Interval Converged - Add Contribution to Running Sums
         /*
         /**/

    Result = Result + QNow;
    ErrEst = ErrEst + EstErr;
    Cor11  = Cor11  + QDiff/1023.0;

    /*  Locate Next Interval  */

    while (Nim != 2*(Nim / 2))
    {
        Nim = Nim / 2;
        Lev--;
    };

    Nim++;

    if (Lev <= 0)  goto L80;

    /*  Assemble Elements Required for the Next Interval  */

    QPrev = QRight [Lev];

    X [0] = X [16];
    F [0] = F [16];

    for (i = 1; i <= 8; i++)
    {
        F [2*i] = FSave [i][Lev];
        X [2*i] = XSave [i][Lev];
    };

    goto L30;


    L80: /**/
         /*
         /*  Stage 8 - Finalize and Return
         /*
         /**/

    Result = Result + Cor11;

    /*  Make Sure ErrEst Not Less Than Roundoff Level  */

    if (ErrEst != 0.0)
    {
        Temp = fabs (Result) + ErrEst;

        while (Temp == fabs (Result))
        {
            ErrEst = 2.0*ErrEst;
            Temp   = fabs (Result) + ErrEst;
        };

    };


    return (Result);

};  /*  double Integrate  */








