
#include "aradengn.h"








             /****    G L O B A L    D A T A    T Y P E S    ****/


typedef struct
        {
            float TG,
                  XG,
                  YG,
                  RG,
                  CosA,
                  SinA,
                  XpRot,
                  YpRot,
                  GaussQ,
                  GaussCoeffX,
                  GaussCoeffY;

            int Bin,
                Division;

        } ParcelRecord;  /*  struct ParcelRecord  */

typedef ParcelRecord ParcelRecordArray [451];








         /****    G L O B A L    V A R I A B L E S    T H A T     ****/
         /****    D E F I N E    A N    A I R R A D    C A S E    ****/


double W;   /*  Total Yield of Burst - kt    */
double FW;  /*  Fission Yield of Burst - kt  */

int FissionType;  /*  A code representing the type of fission reaction.  */
                  /*  0 is a code to make a best guess based on the      */
                  /*  burst yields.                                      */

double HOB;  /*  HOB - Height of burst above ground level - meters      */
double ZGZ;  /*  ZGZ - Mean sea level altitude of ground zero - meters  */


double PRate;  /*  Precipitation rate - mm/hr  */

int NumberOfLayers;  /*  The number of wind measurement heights.  */

WindArrayType WindMeasurementHeight,
              WindSpeed,
              WindDirection;

    /*  WindMeasurementHeight [i] - Height at which a wind measurement i  */
    /*                              was made - meters above ground level  */
    /*  WindSpeed [i]             - Wind speed at this height - m/s       */
    /*  WindDirection [i]         - From direction of wind at this height */
    /*                              - degrees east of north               */


int NumXGridPoints,  /*  Number of grid points in along wind direction  */
    NumYGridPoints;  /*  Ditto in across wind direction                 */

double XGridSpacing,  /*  Distance between adjacent grid points, in  */
       YGridSpacing,  /*  in along- and across-wind directons.       */

       XGridOrigin,   /*  Location of lower left corner of grid, in   */
       YGridOrigin;   /*  meters downwind and acrosswind from ground  */
                      /*  zero.                                       */


int NPS;  /*  Perform calculations using every  */
          /*  NPS'th particle size bin (1-4)    */

int NCD;  /*  Divide cloud into NCD wafers for tracking (1-5)  */

double R_Search;

    /*  Number of sigmas out from center of a puff before the mateial   */
    /*  outside can be considered negligible.  If R_Search > 3.5, then  */
    /*  R_Search is assumed to be infinite. (ie, no RoundInation when     */
    /*  adding Gaussian parcels.) - dimensionless                       */


int NumReceptors;  /*  Number of specific location receptors where  */
                   /*  exact fallout calculations will be made.     */

double ReceptorX [MaxReceptors],
       ReceptorY [MaxReceptors];

    /*  Location of specific location receptor i,    */
    /*  in meters east and north of ground zero.     */
    /*  Coordinates are changed to the rotated wind  */
    /*  coordinate system in ProcessWindData ().     */








  /****    I N I T I A L I Z E D    G L O B A L    V A R I A B L E S    ****/
  /****         C O N T A I N I N G    L I B R A R Y    D A T A         ****/


const double Kr = 1000.0;

    /*  Constant-K horizontal dispersion parameter, m^2/s  */


#define  Total_NPS  75

    /*  Total number of particle size bins  */

double PD [Total_NPS] =
      {4.6189E-3,2.7058E-3,1.9869E-3,1.6103E-3,1.3673E-3,
       1.1937E-3,1.0616E-3,9.5693E-4,8.7138E-4,7.9984E-4,7.3891E-4,
       6.8627E-4,6.4022E-4,5.9953E-4,5.6326E-4,5.3069E-4,5.0125E-4,
       4.7449E-4,4.5003E-4,4.2759E-4,4.0691E-4,3.8777E-4,3.7001E-4,
       3.5348E-4,3.3805E-4,3.2360E-4,3.1005E-4,2.9731E-4,2.8530E-4,
       2.7397E-4,2.6325E-4,2.5310E-4,2.4347E-4,2.3433E-4,2.2562E-4,
       2.1733E-4,2.0943E-4,2.0188E-4,1.9466E-4,1.8775E-4,1.8114E-4,
       1.7480E-4,1.6871E-4,1.6287E-4,1.5725E-4,1.5184E-4,1.4664E-4,
       1.4163E-4,1.3680E-4,1.3213E-4,1.2763E-4,1.2328E-4,1.1908E-4,
       1.1501E-4,1.1107E-4,1.0725E-4,1.0355E-4,9.9963E-5,9.6483E-5,
       9.3105E-5,8.9825E-5,8.6638E-5,8.3541E-5,8.0529E-5,7.7599E-5,
       7.4748E-5,7.1972E-5,6.9268E-5,6.6632E-5,6.4063E-5,6.1558E-5,
       5.9112E-5,5.6725E-5,5.4394E-5,5.2116E-5};

    /*  Particle diameters for the 75 size classifications, meters.  */


double FO [Total_NPS] =
      {18.113,13.661, 11.319,  9.8484, 8.7798, 7.9459, 7.2655, 6.6946,
       6.2047, 5.7777, 5.4005, 5.0641, 4.7612, 4.4867, 4.2363, 4.0067, 3.7952,
       3.5996, 3.4180, 3.2490, 3.0911, 2.9433, 2.8045, 2.6741, 2.5512, 2.4351,
       2.3254, 2.2216, 2.1230, 2.0295, 1.9406, 1.8560, 1.7754, 1.6987, 1.6253,
       1.5553, 1.4884, 1.4244, 1.3631, 1.3044, 1.2482, 1.1943, 1.1426, 1.0930,
       1.0453, 0.99953,0.95559,0.91335,0.87274,0.83359,0.79600,0.75981,0.72501,
       0.69145,0.65912,0.62795,0.59793,0.57242,0.54529,0.51590,0.48642,0.45791,
       0.43340,0.40902,0.38555,0.36297,0.34125,0.32037,0.30031,0.28107,0.26261,
       0.24492,0.22799,0.21180,0.19633};

    /*  Particle Setteling Speed at Sea Level, meters/second.  */


double TFT [9] =
      {300.0, 300.0, 300.0, 300.0, 300.0, 280.0, 200.0, 160.0, 150.0};

    /*  Cloud Stop Times for Kilotonage of 10^I, I = -3, 5,  seconds.  */


double TST [9] =
      {421.0, 421.0, 381.0, 382.0, 422.0, 663.0, 783.0, 787.0, 991.0};

    /*  Cloud Stabilization for Kilotonage of 10^I, I = -3, 5,  seconds.  */


double FK [8] = 
     {0.0,6.3010E9,6.0830E9,6.9733E9,7.2911E9,7.8643E9,7.9407E9,8.2111E9};

    /*  K Activity values for each fission type,  */
    /*  (Roentgens m^2 hr^-1 KT^-1)               */


char FissionTypeName [8][9] = 
      {"Estimate", "U233HE", "P239HE", "P239FI",
       "U235HE", "U235FI", "U238TN", "U238HE"};

    /*  Names of the different fission types.
    /*
    /*  The following symbols are used in the names...
    /*
    /*  U = Uranium     P = Plutonium
    /*
    /*  HE = High Energy Neutron Fission
    /*  FI = Fission Spectrum Neutron Fission
    /*  TN = Thermonuclear Fission
    /*
    /**/


double AA [Total_NPS + 1] =  
     {0.005826,0.005980,0.006088,0.006170,0.006240,0.006303,0.006361,
      0.006415,0.006467,0.006516,0.006564,0.006611,0.006656,0.006701,0.006745,
      0.006788,0.006831,0.006874,0.006916,0.006959,0.007001,0.007043,0.007086,
      0.007128,0.007170,0.007213,0.007256,0.007299,0.007343,0.007387,0.007431,
      0.007456,0.007521,0.007567,0.007613,0.007660,0.007708,0.007756,0.007805,
      0.007854,0.007905,0.007956,0.008009,0.008062,0.008116,0.008172,0.008228,
      0.008286,0.008344,0.008405,0.008466,0.008529,0.008594,0.008660,0.008728,
      0.008798,0.008870,0.008944,0.009020,0.009099,0.009180,0.009263,0.009350,
      0.009439,0.009532,0.009628,0.009728,0.009832,0.009940,0.010053,0.010171,
      0.010294,0.010423,0.010558,0.010701,0.010851};

    /*  Particle Activity Fractions, dimensionless.  */


double B1 = 2.90E-5;
double B2 = 4.05E-5;

    /*  Coeff. for adjustment of particle fallspeed for size of particle.  */








         /****    G L O B A L    V A R I A B L E S    U S E D    ****/
         /****      D U R I N G    C O M P U T A T I O N S       ****/


int ParticleBinIndex;

double FP [Total_NPS];

ParcelRecordArray PR;

int NumParcelRecords;

double RI,
       RF,
       TI,
       TF,
       TS,
       ZBI,
       ZTI,
       ZBF,
       ZTF,
       SCX,
       SCY,
       YBI,
       YTI,
       YTS,
       FarX,
       FarY,
       FarR,
       B1GZ,
       EXB1GZ,
       B2GZ,
       EXB2GZ,
       RCFB,
       ZBTRP,
       ZTTRP,
       Fd,
       Two_Kr,
       CAYF,
       ACNST,
       XCoeff0,
       XCoeff1,
       YCoeff0,
       YCoeff1,
       Alt1Coeff0,
       Alt1Coeff1,
       Alt2Coeff0,
       Alt2Coeff1,
       SqrtCutoffTime,
       ZMB,
       ZMT,
       TMB,
       TMT,
       ZI_Terp_Factor;

WindArrayType Uwind, Vwind, d2dtUwind, d2dtVwind, WindMeasurementAltitude;








                      /****    F U N C T I O N S    ****/


void UpdateRunProgressIndicator (int Stage, float PercentComplete)

/*  This function is called periodically during the computational stage of
/*  the computational engine.  It can be used to drive a progress-indicator
/*  of some sort for the user.  This dummy version does nothing, but the
/*  user can replace it with a version that performs whatever action is
/*  appropriate.
/*
/*  Stage is the current computational stage of the engine, either 1 or 2.
/*  PercentComplete is the approximate percentage of the current stage
/*  that has been completed (0.0 to 100.0).
/*
/**/

{
    #ifdef Debug
        printf ("Stage %i now %.1f %% complete\n", Stage, PercentComplete);
    #endif


    Stage = PercentComplete;   /*  To prevent "parameter not used" warnings  */
    PercentComplete = Stage;   /*  until this function defined by end user   */

};  /*  void UpdateRunProgressIndicator  */








int DecodeFissionType (char *s)

/*  Maps the character representation of the fission type to its
/*  integer code used internally by Airrad.
/*
/**/

{
    int i;

    for (i=0; i<=7; i++)
       if (strstr (s, FissionTypeName [i])) break;

    if (i>7) FailExit ("Unrecognized fission type");

    return (i);

};  /*  int DecodeFissionType  */








void ReadCaseVariables (char *FileName)

/*  Reads the ASCII Airrad case description file named FileName, and
/*  stores the data internally, to be used by the next call to
/*  PerformComputations.
/*
/**/

{
    FILE* f;

    char s [128];

    double Zmod;  /*  Used to convert met profile Z's to heights  */

    double X1, X2, Y1, Y2;  /*  Coordinates of lower-left and upper-right  */
                            /*  corners of grid, in meters alongwind and   */
                            /*  across-wind from the burst locaton.        */

    int i;



    f = fopen (FileName, "r");

    if (f == NULL) FailExit ("Could not open case description file");



    fgets (s, 128, f);  /*  Skip #Title line  */

    fgets (s, 128, f);  /*  Read case title  */

    #ifdef Debug
      printf ("Reading Case: %s\n", s);
    #endif



    fgets (s, 128, f);  /*  Skip #Burst line  */

    fscanf (f, " %lf %lf %lf %s ", &HOB,
                                   &W,
                                   &FW,
                                    s);

    FissionType = DecodeFissionType (s);

    #ifdef Debug
      printf ("\nHOB = %lf, W = %lf, FW = %lf, FissionType = %i (%s)\n",
               HOB, W, FW, FissionType, s);
    #endif



    fgets (s, 128, f);  /*  Skip #Meteorology line  */

    fscanf (f, " %lf %i %s %lf ", &PRate,&NumberOfLayers, s, &ZGZ);

    if (strstr (s, "msl"))
        Zmod = -ZGZ;
    else
        Zmod = 0.0;

    #ifdef Debug
      printf ("\nPR = %lf, n = %i, Z's are in = %s, Zmod = %lf, ZGZ = %lf\n",
               PRate, NumberOfLayers, s, Zmod, ZGZ);
    #endif


    for (i = 0; i < NumberOfLayers; i++)
    {
        fscanf (f, " %lf %lf %lf ", &WindMeasurementHeight [i],
                                    &WindDirection [i],
                                    &WindSpeed [i]);

        WindMeasurementHeight [i] += Zmod;


        #ifdef Debug
          printf ("   Level %i, Height = %lf, Direction = %lf, Speed = %lf\n",
                   i+1, WindMeasurementHeight [i],
                        WindDirection [i],
                        WindSpeed [i]);
        #endif

    };



    fgets (s, 128, f);  /*  Skip #Grid line  */

    fscanf (f, " %i %i %lf %lf %lf %lf ", &NumXGridPoints,
                                          &NumYGridPoints,
                                          &X1, &X2, &Y1, &Y2);

    XGridOrigin = X1;
    YGridOrigin = Y1;

    XGridSpacing = (X2 - X1) / (NumXGridPoints-1);
    YGridSpacing = (Y2 - Y1) / (NumYGridPoints-1);

    #ifdef Debug

      printf ("\nNumXGridPoints = %i, NumYGridPoints = %i\n",
               NumXGridPoints, NumYGridPoints);

      printf ("\nX1 = %lf, X2 = %lf, Y1 = %lf, Y2 = %lf\n",
               X1, X2, Y1, Y2);

      printf ("XGridSpacing = %lf, YGridSpacing = %lf\n",
               XGridSpacing, YGridSpacing);

      printf ("XGridOrigin = %lf, YGridOrigin = %lf\n",
               XGridOrigin, YGridOrigin);
               
    #endif



    fgets (s, 128, f);  /*  Skip #Model line  */

    fscanf (f, " %lf %i %i ", &R_Search, &NPS, &NCD);

    if (R_Search < 0.1) R_Search = 9.9e99;

    #ifdef Debug
      printf ("\nR_Search = %lf, NPS = %i, NCD = %i\n",
               R_Search, NPS, NCD);
    #endif



    fgets (s, 128, f);  /*  Skip #Receptors line  */

    fscanf (f, " %i ", &NumReceptors);

    #ifdef Debug
      printf ("\nNumReceptors = %i\n", NumReceptors);
    #endif


    for (i=0; i<NumReceptors; i++)
    {
        fscanf (f, " %lf %lf ", &ReceptorX [i],
                                &ReceptorY [i]);

        #ifdef Debug
          printf ("   Receptor %i, X = %lf, Y = %lf\n",
                   i+1, ReceptorX [i], ReceptorY [i]);
        #endif

    };



    fclose (f);

};  /*  void ReadCaseVariables  */








void ComputeBurstConstants (void
)

/*   Initializes constants for a given burst   */

{

    int Default_Fission_Type_Low_Yield  = 2;  /*  Less than 1000 kt     */
    int Default_Fission_Type_High_Yield = 6;  /*  Greater then 1000 kt  */

    double Terp, Ylog, Zscl;

    int IY;


    if (FissionType == 0)
    {

        if (W < 1000.0)
           FissionType = Default_Fission_Type_Low_Yield;
        else
          FissionType = Default_Fission_Type_High_Yield;

        #ifdef Debug
         printf ("\n");
         printf ("Best guess at fission type: %s \n",
                  FissionTypeName [FissionType]);
        #endif

    }
    else
    {
        #ifdef Debug
          printf ("Fission type              : %s \n",
                   FissionTypeName [FissionType]);
        #endif
    };


    CAYF = FK [FissionType];



               /****  Compute initial (fireball) and       ****/
               /****  final (stabilized) cloud properties  ****/


    /*  Construct Interpolation Parameters  */

    Ylog = log10 (W);
    IY   = RoundIn (Ylog) + 3;
    Terp = fabs (Ylog - RoundIn (Ylog));


    /*  Compute initial time, time cloud stops rising, time cloud stabilizes  */

    TI = 2.07*pow (W, 0.19);

    TF = TFT [IY] + Terp*(TFT [IY+1] - TFT [IY]);
    TS = TST [IY] + Terp*(TST [IY+1] - TST [IY]);


    /*  Compute initial and final cloud radii  */

    RI = 108.0 * pow (W, 0.33);
    RF = exp(6.7553 + Ylog*(0.7381 + 0.060308*Ylog));


    /*  Compute initial cloud base and top altitudes  */

    ZBI = ZGZ + 90.0*pow(W, (1.0/3.0)) - 0.66144*RI + HOB;
    ZTI = ZBI + 1.32288*RI;


    /*  Compute final cloud base and top altitudes  */

    if (W > 4.07)
        ZBF = ZGZ + 2661.0*pow (W, 0.2198) + HOB;
    else
        ZBF = ZGZ + 2228.0*pow (W, 0.3463) + HOB;

    if (W <= 2.29)
        ZTF = ZGZ + 3597.0*pow(W, 0.2553) + HOB;
    else if (W <= 19.0)
             ZTF = ZGZ + 3170.0*pow(W, 0.4077) + HOB;
         else
             ZTF = ZGZ + 6474.0*pow(W, 0.1650) + HOB;


    /*  Compute scaled height of burst  */

    Zscl = (HOB*3.2808)/pow (W, (1.0/3.0));


    /*  Compute scaling factor to account for height of burst above ground  */

    if (Zscl > 1300.0)  /*  Simfic recommends aborting     */
                        /*  calculation when Zscl > 180.0  */

        FailExit ("Scaled height too large - No fallout");

    else if (Zscl > 0.0)
             Fd = pow (0.45345, (Zscl/65.0));
         else
             Fd = 1.0;

    Two_Kr = 2.0*Kr;

    ZI_Terp_Factor = (ZTI - ZBI)/NCD;

    SCX = sqrt (TF) - sqrt (TI);

    SCY = ZBF - ZBI;

    YBI = ZBI/SCY;
    YTI = ZTI/SCY;
    YTS = ZTF/SCY;

    ACNST  = (YTI-YBI)/(YTS-YTI-1.0);

    B1GZ   = B1*ZGZ;
    EXB1GZ = exp (B1GZ);

    B2GZ   = B2*ZGZ;
    EXB2GZ = exp (B2GZ);

    RCFB   = RF - RI;

    ZBTRP  = (ZBF - ZBI)/Total_NPS;
    ZTTRP  = (ZTF - ZTI)/Total_NPS;


    #ifdef Debug

      printf ("Associated K Factor       : %13lE (R M^2)/(Hr Kt) \n", CAYF);
      printf ("\n");
      printf ("     Computed Cloud Properties\n");
      printf ("\n");
      printf ("Fireball Definition Time     : %9.2lf s \n", TI);
      printf ("Time When Cloud Stops Rising : %9.2lf s \n", TF);
      printf ("Cloud Stabilization Time     : %9.2lf s \n", TS);
      printf ("\n");
      printf ("Initial Cloud Radius            : %9.2lf m \n", RI);
      printf ("Final Stabilized Cloud Radius   : %9.2lf m \n", RF);
      printf ("\n");
      printf ("Initial Cloud Base Altitude : %9.2lf m \n", ZBI);
      printf ("Initial Cloud Top Altitude  : %9.2lf m \n", ZTI);
      printf ("\n");
      printf ("Final Stabilized Cloud Base Altitude : %9.2lf m \n", ZBF);
      printf ("Final Stabilized Cloud Top Altitude  : %9.2lf m \n", ZTF);
      printf ("\n");
      printf ("Scaled Height of Burst : %9.2lf ft/KT^1/3 \n", Zscl);
      printf ("\n");
      printf ("Burst Height Scaling Factor : %7.5lf \n", Fd);
      printf ("\n");

    #endif

};  /*  void ComputeBurstConstants  */








double XWindVel (double Altitude)

/*  Returns the X (east) component of the wind
/*  velocity, in m/s, at Altitude, in meters.
/*
/**/

{
    #ifdef WindSpline

        if (Altitude <= WindMeasurementAltitude [0])
            return (Uwind [0]);
        else
            if (Altitude >= WindMeasurementAltitude [NumberOfLayers - 1])
                 return (Uwind [NumberOfLayers - 1]);
             else
                 return (Spline (WindMeasurementAltitude,
                                 Uwind, d2dtUwind, NumberOfLayers,
                                 Altitude));

    #else

        return (LinearInterpolate (WindMeasurementAltitude,
                                   Uwind, NumberOfLayers, Altitude));

    #endif

};  /*  double XWindVel  */








double YWindVel (double Altitude)

/*   Returns the Y (north) component of the wind
/*   velocity, in m/s, at Altitude, in meters.
/*
/**/

{
    #ifdef WindSpline

        if (Altitude <= WindMeasurementAltitude [0])
            return (Vwind [0]);
        else
            if (Altitude >= WindMeasurementAltitude [NumberOfLayers - 1])
                return (Vwind [NumberOfLayers - 1]);
            else
                return (Spline (WindMeasurementAltitude,
                                Vwind, d2dtVwind, NumberOfLayers,
                                Altitude));

    #else

        return (LinearInterpolate (WindMeasurementAltitude,
                                   Vwind, NumberOfLayers, Altitude));

    #endif

};  /*  double YWindVel  */








void Interpolate (WindArrayType V, int NumberOfLayers)

/*  Returns V with any undefined elements (ie, = 999) set
/*  by linear interpolation if possible.  Note that this subroutine
/*  does not do extrapolation of undefined elements at each
/*  end of the array.
/*
/**/

{
    int I, LowI, HighI;

    double DeltaV, DeltaHeight;


    HighI = 0;

    do
    {
        LowI = HighI - 1;

        do
            LowI++;
        while ((LowI < (NumberOfLayers-1)) &&
               (MetUndefined (V [LowI]) || ! MetUndefined (V [LowI+1])));

        HighI = LowI + 1;

        do
            HighI++;
        while ((HighI < NumberOfLayers) && MetUndefined (V [HighI]));

        if (HighI < NumberOfLayers)
        {
            I = LowI + 1;

            DeltaHeight = WindMeasurementHeight [HighI] -
                          WindMeasurementHeight [ LowI];

            DeltaV = V [HighI] - V [LowI];

            while (I < HighI)
            {
                V [I] = V [LowI] + DeltaV*(WindMeasurementHeight [I] -
                        WindMeasurementHeight [LowI])/DeltaHeight;

                I++;
            };
        };
    }
    while (HighI < NumberOfLayers);

};  /*  void Interpolate  */








void CheckDirections (WindArrayType V, int N)

/*  V is assumed to be an array of N directions in degrees.
/*  Adds or subtracts multiples of 360 degrees to elements of
/*  V so that no two adjacent elements differ by greater than
/*  180 degrees.  Undefined elements (ie, = 999) are ignored.
/*
/*  This subroutine is called before linear interpolation of undefined
/*  elements is done on V.  For example, if V = (30.0, 999, 350.0),
/*  CheckDirections would return V = (30.0, 999, -10.0), and
/*  procedure Interpolate would return V with the missing value
/*  correctly set to 10.0 instead of 190.0.
/*
/**/

{
    int I, ILast;

    double C, DeltaV, Correction;


    Correction = 0.0;
    ILast      = 0;

    for (I = 1; I < N; I++)
        if (! MetUndefined (V [I]))
        {
            if (! MetUndefined (V [ILast]))
            {
                V [I] += Correction;

                do
                {
                    DeltaV = V [I] - V [ILast];

                    if (fabs (DeltaV) > 180.0)
                    {
                        if (DeltaV > 0.0)
                            C = -360.0;
                        else
                            C =  360.0;

                        V [I]      +=  C;
                        Correction +=  C;

                    };

                } while (fabs (DeltaV) > 180.0);

            };

        ILast = I;

    };

};  /*  void CheckDirections  */








void ProcessWindData (void)

/*  1. Sorts wind data be height.
/*  2. Adds new wind measurement levels to properly represent profile.
/*  3. Interpolates or extrapolates missing data.
/*  4. Transforms heights into altitudes.
/*  5. Transforms wind speed and direction into U, V vector form.
/*  6. Sets up arrays of coefficients so that spline interpolation can
/*     be used to get wind velocity at any arbitrary altitude.
/*
/*  General algorithm for estimating undefined values.
/*
/*     1. Discard any levels with undefined heights.
/*     2. Linearly interpolate any undefined levels between defined levels.
/*     3. for extrapolation of wind speed...
/*
/*           In boundary layer, assume speed with height follows form
/*           Vz == Ccoeff*Z^Bcoeff, where Bcoeff == 0.0423 and Ccoeff can
/*           be found from Ccoeff=Vz/Z^Bcoeff at a known level.
/*
/*           for 300 mb (9000 m.) and above, use table given by arrays
/*           ExtraLevel and ExtraLevelSpeed below.  This is an average
/*           mid-latitude wind profile for the upper troposphere and
/*           stratosphere.
/*
/*     4. for extrapolation of Wind Direction...
/*
/*           In boundary layer, give 19 degree veer.
/*
/*           for 300 mb (9000 m.) and above, assume wind is from west
/*           (270 degrees).
/*
/*
/**/

{
    #define  NumExtraLevels  9

    double ExtraLevel [NumExtraLevels] =
              {2.0, 100.0, 1000.0, 9000.0, 12000.0,
               13500.0, 16000.0, 20000.0, 30000.0};

    double ExtraLevelSpeed [NumExtraLevels] =
               {0.0, 0.0, 0.0, 12.3, 15.7, 13.2, 5.8, 10.0, 20.0};

    double dDIRdZ = 0.019;   /*  Rate of veer in boundary layer, degrees/meter.  */

    double Bcoeff = 0.0423;   /*  Wind velocity in boundary layer: Vz == Ccoeff * Z^Bcoeff,  */
                        /*  where Ccoeff == Vz/Z^Bcoeff at a known layer.              */

    double VeryLargeValue = 9.9e29;


    int i, j, IMaxDir, IMinDir, IMaxSpd, IMinSpd, NumUndefinedHeights;

    double Ccoeff;

    int Found;

    int iZTF;

    double sumux, sumuy, sinrot, cosrot, x, y;



    /*  Count layers with undefined heights - make sure they will appear at  */
    /*  top of array by setting heights to very large values before sort.    */

    NumUndefinedHeights = 0;

    for (i = 0; i < NumberOfLayers; i++)
        if (MetUndefined (WindMeasurementHeight [i]))
        {
            NumUndefinedHeights++;
            WindMeasurementHeight [i] = VeryLargeValue;
        };


    /*  Add new levels to properly represent wind profile  */

    for (i = 0; i < NumExtraLevels; i++)
    {
        Found = 0;

        for (j = 0; j < NumberOfLayers; j++)
            Found = Found || (WindMeasurementHeight [j] == ExtraLevel [i]);

        if (! Found)
        {
            WindMeasurementHeight [NumberOfLayers] = ExtraLevel [i];
            WindDirection         [NumberOfLayers] = 999;
            WindSpeed             [NumberOfLayers] = 999;

            NumberOfLayers++;
        };
    };


    /*  Sort new levels into place.  Levels with    */
    /*  undefined heights will go to top of array.  */

    HeapSort (WindMeasurementHeight, NumberOfLayers, 1);
    HeapSort (WindSpeed,             NumberOfLayers, 2);
    HeapSort (WindDirection,         NumberOfLayers, 3);


    /*  Get rid of any layers with undefined heights.  */

    NumberOfLayers = NumberOfLayers - NumUndefinedHeights;


    /*  Check for two wind vectors at same level.  */

    for (i = 0; i < (NumberOfLayers-1); i++)
        if (WindMeasurementHeight [i] == WindMeasurementHeight [i+1])
            FailExit ("Two wind vectors specified at same height");


    /*  Avoid 0/360 degree discontinuity interpolation problem.  */

    CheckDirections (WindDirection, NumberOfLayers);


    /*  Interpolate any missing wind speeds or  */
    /*  directions enclosed by known values.    */

    Interpolate (WindDirection, NumberOfLayers);
    Interpolate (WindSpeed,     NumberOfLayers);


    /*  Find index to highest and lowest height  */
    /*  at which speed and direction are given.  */

    IMaxDir = -1;
    IMinDir = -1;
    IMaxSpd = -1;
    IMinSpd = -1;

    for (i = 0; i < NumberOfLayers; i++)
    {
        if (! MetUndefined (WindDirection [i]))
        {
            IMaxDir = i;
            if (IMinDir == -1)  IMinDir = i;
        };

        if (! MetUndefined (WindSpeed [i]))
        {
            IMaxSpd = i;
            if (IMinSpd == -1)  IMinSpd = i;
        };
    };


    /*  Estimate unknown wind directions  */

    if (IMinDir == -1)
    {

        IMinDir = NumberOfLayers - 1;
        IMaxDir = NumberOfLayers - 1;

        WindDirection [IMinDir] = 270.0;

            /*  If you know absolutely nothing, assume west wind.  */

    };


    /*  Estimate Wind Direction Below Lowest Known Layer  */

    for (i = IMinDir-1; i >= 0; i--)
    {
        if (WindMeasurementHeight [i] >= 1000.0)
            WindDirection [i] = WindDirection [IMinDir];

               /*  if above boundary layer, assume constant.  */

        else

            WindDirection [i] = WindDirection [IMinDir] -
                 dDIRdZ * (Min (1000.0, WindMeasurementHeight [IMinDir]) -
                           WindMeasurementHeight [i]);

                /*  If in boundary layer, give 19 degree veer  */
    };


    /*  Estimate Wind Direction Above Highest Known Layer  */

    for (i = IMaxDir+1; i < NumberOfLayers; i++)
    {
        if (WindMeasurementHeight [i] >= 9000.0)

            WindDirection [i] = 270.0;

                /*  If above ~ 300 mb, assume west wind  */

        else if (WindMeasurementHeight [i] > 1000.0)

                 WindDirection [i] = WindDirection [i - 1];

                    /*  If in layer between boundary layer  */
                    /*  and ~ 300 mb, assume constant       */

             else

                 WindDirection [i] = WindDirection [IMaxDir] +
                                   dDIRdZ*(WindMeasurementHeight [i] -
                                           WindMeasurementHeight [IMaxDir]);

                     /*  If in boundary layer (top at ~ 1000 m.),  */
                     /*  give 19 degree veer.                      */
    };


    /*  Map wind directions back to 0-360 range.  */

    for (i = 0; i < NumberOfLayers; i++)
        WindDirection [i] = ((long int) Round (WindDirection [i])) % 360;


    /*  Estimate unknown wind speeds  */


    /*  Set certain high ( > 300 mb) levels to average mid-latitude values  */

    for (i = IMaxSpd + 1; i < NumberOfLayers; i++)
        for (j = 3; j < NumExtraLevels; j++)
            if (WindMeasurementHeight [i] == ExtraLevel [j])
            {
                if (IMinSpd == -1) IMinSpd = i;
                WindSpeed [i] = ExtraLevelSpeed [j];
            };

    if (IMaxSpd == -1) IMaxSpd = NumberOfLayers - 1;



    /*  Estimate Wind Speed Below Lowest Known Layer  */

    Ccoeff =                 WindSpeed [IMinSpd]  /
             pow (Min (1000.0, WindMeasurementHeight [IMinSpd]), Bcoeff);

    for (i = IMinSpd - 1; i >= 0; i--)
        if (WindMeasurementHeight [i] >= 1000.0)

            WindSpeed [i] = WindSpeed [IMinSpd];

                /*  If above boundary layer, assume constant.  */

        else

            WindSpeed [i] = Ccoeff * pow (WindMeasurementHeight [i], Bcoeff);

                /*  if in boundary layer, give pow decay of velocity  */


    /*  Estimate Wind Speed Above Highest Known Layer  */

    Ccoeff =         WindSpeed [IMaxSpd]  /
             pow (WindMeasurementHeight [IMaxSpd], Bcoeff);

    for (i = IMaxSpd + 1; i < NumberOfLayers; i++)
    {

        if (WindMeasurementHeight [i] >= 9000.0)

            /* Do nothing */ ;

                /*  If above 300 mb, has already been set above  */


        else if (WindMeasurementHeight [i] > 1000.0)

                 WindSpeed [i] = WindSpeed [i - 1];

                     /*  If in layer between boundary layer  */
                     /*  and ~ 300 mb, assume constant       */

             else

                 WindSpeed [i] = Ccoeff*pow (WindMeasurementHeight [i], Bcoeff);

                     /*  If in boundary layer (top at 1000 m.),  */
                     /*  give pow rise of velocity               */
    };


    /*  Fill in any undefined speed between average speed layers  */

    Interpolate (WindSpeed, NumberOfLayers);



    /*  Transform Direction - Speed wind pairs into U - V vector pairs.  */

    for (i = 0; i < NumberOfLayers; i++)
    {
        WindMeasurementAltitude [i] = WindMeasurementHeight [i] + ZGZ;

        Uwind [i] = -WindSpeed [i] * cos (HalfPi - WindDirection [i]*DegToRad);
        Vwind [i] = -WindSpeed [i] * sin (HalfPi - WindDirection [i]*DegToRad);
    };


    #ifdef WindSpline

        /*  Set up for spline interpolation of wind U and V with altitude.  */

        NewSpline (WindMeasurementAltitude, Uwind, NumberOfLayers,
                   1.0e30, 1.0e30, d2dtUwind);

        NewSpline (WindMeasurementAltitude, Vwind, NumberOfLayers,
                   1.0e30, 1.0e30, d2dtVwind);

    #endif


    #ifdef Debug

        printf ("\n");
        printf ("\n");
        printf ("               Processed Wind Data \n");
        printf ("\n");
        printf (" Layer    Altitude (m)   Speed (m/s)   "
                "Direction (deg)    U (m/s)   V (m/s) \n");
        printf ("\n");

        for (i = 0; i < NumberOfLayers; i++)
            printf ("%5i%13.0lf%15.2lf%15.0lf%16.2lf%10.2lf \n",
                     i + 1, WindMeasurementAltitude [i],
                     WindSpeed [i], WindDirection [i],
                     Uwind [i], Vwind [i]);

    #endif



    /*  If don't want to rotate the computational grid so that its X axis  */
    /*  is parallel to the mean wind direction, then return here.          */

    #ifdef DoNotRotateGrid

        return;

    #endif



    /*  Compute angle to rotate wind coordinate system by so that the  */
    /*  X axis of the computational grid is parallel to the mean wind  */
    /*  vector between the ground and the top of the stabilized cloud  */
    /*  ZTF.                                                           */

    /*  First, determine iZTF such that WindMeasurementAltitufe [iZTF] <=  */
    /*  ZTF < WindMeasurementAltitude [iATF+1]                             */

    iZTF = 0;

    Hunt (WindMeasurementAltitude, NumberOfLayers, ZTF, &iZTF);


    /*  Sum all u and v componenets of the wind layers from ground to ZTF  */

    sumux = 0.0;
    sumuy = 0.0;

    for (i=0; i <= Max (0, iZTF); i++)
    {
        sumux += Uwind [i];
        sumuy += Vwind [i];
    };


    /*  Determine the angle of the summed components vector.  */

    if (fabs (sumux) > 1.0e-6)
    {
        GridRotationAngle = atan (sumuy/sumux);
        if (sumux < 0.0) GridRotationAngle += Pi;
    }
    else if (sumuy < 0.0)
             GridRotationAngle = -0.5*Pi;
         else
             GridRotationAngle =  0.5*Pi;

    /*  Rotate each wind vector so that the mean wind direction is now  */
    /*  pointed along the X axis of the computational grid.             */

    cosrot = cos (GridRotationAngle);
    sinrot = sin (GridRotationAngle);

    for (i=0; i < NumberOfLayers; i++)
    {
        x = Uwind [i];
        y = Vwind [i];

        Uwind [i] = x*cosrot + y*sinrot;
        Vwind [i] = y*cosrot - x*sinrot;
    };


    /*  Transform the coordinates of the specific receptor locations from    */
    /*  (X,Y) in east-north space to (X-Y) in along-wind, across-wind space  */

    for (i=0; i < NumReceptors; i++)
    {
        x = ReceptorX [i];
        y = ReceptorY [i];

        ReceptorX [i] = x*cosrot + y*sinrot;
        ReceptorY [i] = y*cosrot - x*sinrot;
    };



    #ifdef Debug

        printf ("\n");
        printf ("\n");
        printf ("Grid Rotation Information \n");
        printf ("\n");
        printf ("Mean wind vector from ground to top of cloud stabilization\n");
        printf ("altitude %9.2lf m is (%7.2lf, %7.2lf) \n", ZTF, sumux, sumuy);
        printf ("\n");
        printf ("Computed Grid Rotation Angle = %6.2lf degrees \n",
                 GridRotationAngle*RadToDeg);
        printf ("\n");
        printf ("               Rotated Wind Data \n");
        printf ("\n");
        printf (" Layer    Altitude (m)      U (m/s)   V (m/s) \n");
        printf ("\n");

        for (i = 0; i < NumberOfLayers; i++)
            printf ("%5i%13.0lf%16.2lf%10.2lf \n",
                     i + 1, WindMeasurementAltitude [i],
                     Uwind [i], Vwind [i]);

        printf ("\n");
        printf (" Specific Receptors Locations in Rotated Coordinates\n");
        printf ("\n");
        printf ("  Receptor Name        X Location    Y Location \n");
        printf ("\n");

        for (i=0; i < NumReceptors; i++)
            printf ("   Receptor %2i        %8.0lf m    %8.0lf m \n",
                     i + 1, ReceptorX [i], ReceptorY [i]);

    #endif

};  /*  void ProcessWindData  */








double Average_Fallspeed (double Altitude, int BinIndex)

/*   Returns the average fallspeed of a particle in particle size bin
/*   BinIndex, through the layer from Altitude to ZGZ (Ground).
/*   All units are MKS.
/*
/**/

{
    const double Large_Particle_Limit = 3.0E-4;

    double Dummy;


    if (PD [BinIndex] < Large_Particle_Limit)
    {
        Dummy = B1*Altitude;

        return (FO [BinIndex] * (exp (Dummy) - EXB1GZ)/(Dummy - B1GZ));
    }
    else
    {
        Dummy = B2*Altitude;

        return (FO [BinIndex] * (exp (Dummy) - EXB2GZ)/(Dummy - B2GZ));
    };

};  /*  double Average_Fallspeed  */








void Compute_Activity_Fraction_Array (void)

/*   Computes the array FP [i], where i is a particle size index.
/*   FP [i] is the total amount of activity associated with particles of size
/*   class i.
/*
/**/

{
    double Activity;

    int i;


    Activity = FW*Fd*CAYF / NCD;


    switch (NPS)
    {

    case 1: for (i=0; i < Total_NPS; i++)
                FP[i] = Activity*AA[i];

            break;


    case 2: FP[0] = Activity*(AA[0] + AA[1]/2.0);

            for (i=2; i < Total_NPS; i += 2)
                FP[i] = Activity*(AA[i] + (AA[i-1] + AA[i+1])/2.0);

            break;


    case 3: FP[0] = Activity*(AA[0] + AA[1]);

            for (i=3; i < Total_NPS; i += 3)
                FP[i] = Activity*(AA[i] + AA[i-1] + AA[i+1]);

            break;


    case 4: FP[0] = Activity*(AA[0] + AA[1] + AA[2]/2.0);

            for (i=4; i < Total_NPS; i += 4)
                FP[i] = Activity*(AA[i] + AA[i-1] + AA[i+1] +
                                  (AA[i-2] + AA[i+2])/2.0);

            break;

    };    /*  switch (NPS)   */

};  /*  void Compute_Activity_Fraction_Array  */








double FNEWTN (double X, double B, double C)

/*  Sub-function used during newton iterations as part of  */
/*  function ComputeMaximumHeightAndTimeReached.           */

{

    return (-X*(X + B*log (X) - C)/(X+B));

};  /*  double FNEWTN  */








void ComputeMaximumHeightAndTimeReached
         (int BinIndex, char CloudEnd, double* ZM, double* TM)

/*  Input the particle size bin index BinIndex (0..Total_NPS-1) and CloudEnd,
/*  which may be either 'T' for a particle initially at the cloud top, or
/*  'B' for a particle initially at the cloud base.
/*
/*  Computes maximum particle altitude, ZM, and time when this is
/*  reached, TM, by a method based on the observation that cloud
/*  height increases approximately as the square root of time.
/*
/**/

{
    const double Iteration_Cutoff = 1.0E-5;

    double BXM10,
           BXM13,
           BXO,
           C,
           DA,
           DAO,
           DB,
           DBO,
           DXO,
           F,
           FF,
           PSI,
           PSIM,
           XF,
           XI,
           XM,
           XMO,
           XO,
           YI,
           YL,
           YM,
           YMO,
           YO,
           ZI,
           ZT;

    int IXO,
        IXM10,
        IXM13;

    int Done,
        Do_Iteration_One,
        Do_Iteration_Two,
        Do_Iteration_Three,
        JSKP;


    switch (CloudEnd)
    {

    case 'T':
    case 't': ZI  = ZTI;
           *ZM = ZTI + BinIndex*ZTTRP;

           break;

    case 'B':
    case 'b': ZI  = ZBI;
           *ZM = ZBI + BinIndex*ZBTRP;

           break;

    };    /*  switch (CloudEnd)...  */


    do
    {
        ZT = *ZM;

        FF = Average_Fallspeed (ZT, BinIndex);

        YI = ZI/SCY;
        XI = sqrt (TI)/SCX;
        XF = sqrt (TF)/SCX;

        BXO = XI - ACNST;
        DA  = XI - ACNST;
        DB  = XI - YBI;

        DAO = (YI - YBI + ACNST)/ACNST;

        BXM13 = (XI - YBI)/2.0;

        BXM10 = BXO/2.0;

        F = FF*SCX*SCX/SCY;

        IXO   = 0;
        IXM13 = 0;
        IXM10 = 0;

        JSKP = 0;
        Done = 0;

        if (YI != YBI)
        {
            Do_Iteration_One = 1;
        }
        else
        {
            Do_Iteration_One   = 0;
            Do_Iteration_Two   = 1;
            Do_Iteration_Three = 0;
            XO                 = XI;
            YO                 = YI;
        };


        if (Do_Iteration_One)
        {

            /*  Compute time of particle fallout    */
            /*  from cloud cap by newton iteration  */

            C   = (YI - YBI)/(2.0*ACNST*F) + BXO*log (ACNST) + ACNST;
            PSI = C;

            do
            {
                PSIM = PSI;
                PSI  = PSIM + FNEWTN (PSIM, BXO, C);
                IXO  = IXO  + 1;
            }
            while (fabs ((PSI - PSIM)/PSI) >= Iteration_Cutoff);

            XO = PSI + BXO;
            YO =  XO -  XI + YBI;

            if (XO <= XF)
            {
                Do_Iteration_Two = 1;
            }
            else
            {

                Do_Iteration_Two   = 0;
                Do_Iteration_Three = 1;
                JSKP               = 1;

                YL = YI + DAO*(XF - XI) - 2.0*F*(XF - DA)*
                     (XF - XI + DA*log ((XF - DA)/ACNST));

            };
        };


        if (Do_Iteration_Two)
        {

            /*  Compute time at maximum height by newton  */
            /*  iteration for below-cloud particle        */

            DXO = XO - DB;
            DBO = YO / DXO;

            C = YO/(XO - XI + YBI)/(4.0*F) + XO/2.0 +
                BXM13 * log (XO - XI + YBI) - XI + YBI;

            PSI = C;

            do
            {
                PSIM = PSI;

                if (PSIM <= 0.0) PSIM = YBI;

                PSI   = PSIM + FNEWTN (PSIM, BXM13, C);
                IXM13 = IXM13 + 1;
            }
            while (fabs ((PSI - PSIM)/PSI) >= Iteration_Cutoff);

            XM = PSI + XI - YBI;
            YM = 2.0*XM*F*(XM - XI + YBI);

        };


        if (Do_Iteration_Three)
        {

            /*  Compute time at maximum height by newton  */
            /*  iteration for in-cloud particle           */

            C = (YI - YBI + ACNST)/(4.0*F*ACNST) +
                 BXM10*log (ACNST) + XI/2.0 + ACNST;

            PSI = C;

            do
            {
                PSIM = PSI;

                if (PSIM <= 0.0) PSIM = ACNST;

                PSI   = PSIM  + FNEWTN (PSIM, BXM10, C);
                IXM10 = IXM10 + 1;
            }
            while (fabs((PSI - PSIM)/PSI) >= Iteration_Cutoff);

            XMO = PSI - BXO;
            YMO = YBI - ACNST + 2.0*XMO*F*PSI;

            if (JSKP)
            {

                XM = XMO;
                YM = YMO;

                if (XM > XF)
                {
                    XM = XF;
                    YM = YL;
                };

                Done = 1;
            }
            else
            {
                Done = 0;

                if (! ((XM > XO)  &&  (YM > YMO)  ||  (XMO > XO)) )
                {
                    XM = XMO;
                    YM = YMO;
                };
            };

        };


        if (! (Done  ||  (XM <= XF)) )
        {
            XM = XF;

            if (XO > XF) YM = YL;

            if (XO < XF) YM = YO + DBO*(XF - XO) - 2.0*F*(XF - DB)*
                              (XF - XO + DB*log ((XF - DB)/DXO));
        };


        *ZM = YM*SCY;
        *TM = XM*SCX*XM*SCX;

    }
    while (fabs (((*ZM) - ZT)/(*ZM)) > 1.0E-4);


    return;

};  /*  void ComputeMaximumHeightAndTimeReached  */








double XWindFunctionOfTime (double Time)

{
    double Altitude;


    if (Time < SqrtCutoffTime)
        Altitude = Alt1Coeff0 + Alt1Coeff1*sqrt (Time);
    else
        Altitude = Alt2Coeff0 + Alt2Coeff1*Time;

    return (XWindVel (Altitude));

};  /*  double XWindFunctionOfTime  */








double YWindFunctionOfTime (double Time)

{
    double Altitude;


    if (Time < SqrtCutoffTime)
        Altitude = Alt1Coeff0 + Alt1Coeff1*sqrt (Time);
    else
        Altitude = Alt2Coeff0 + Alt2Coeff1*Time;

    return (YWindVel (Altitude));

};  /*  double YWindFunctionOfTime  */








void Compute_Ground_Parameters_For_Each_Parcel (void)

{

    double C,
           RG,
           TG,
           RS,
           TM,
           XG,
           YG,
           ZM,
           ZI,
           ZS,
           ZMin,
           TerpR,
           SHPow,
           SigPow,
           FallSpeed;

    int i, PRIndex;


    #ifdef Debug

        printf ("\n");
        printf ("\n");
        printf ("COMPUTATIONAL  PHASE  ONE\n");
        printf ("\n");
        printf ("Compute a list of the maximum heights and times \n");
        printf ("reached for top and bottom of every parcel. \n");
        printf ("\n");

    #endif



    PRIndex          = 0;
    ParticleBinIndex = 0;

    UpdateRunProgressIndicator (1, 0.0);

    do
    {

        ComputeMaximumHeightAndTimeReached (ParticleBinIndex, 'T', &ZMT, &TMT);
        ComputeMaximumHeightAndTimeReached (ParticleBinIndex, 'B', &ZMB, &TMB);

        #ifdef Debug

            printf ("\n");
            printf ("\n");
            printf ("Particle Size Bin : %2i \n", ParticleBinIndex + 1);
            printf ("\n");
            printf ("ZMB, ZMT = %8.2lf m.   %8.2lf m. \n", ZMB, ZMT);
            printf ("TMB, TMT = %8.2lf 2.   %8.2lf s. \n", TMB, TMT);

        #endif


        if (TMB >= TS)
            ZMin = ZMB;
        else
            ZMin = ZMB - (TS - TMB)*Average_Fallspeed (ZMB, ParticleBinIndex);

        TerpR = RCFB / (ZBF - ZMin);


        for (i = NCD; i >= 0; i--)
        {
            ZI = ZBI + ZI_Terp_Factor*i;

            TM = TMB + pow ((1.0*i)/(1.0*NCD), 0.85)*(TMT - TMB);
            ZM = ZMB + pow ((1.0*i)/(1.0*NCD), 0.85)*(ZMT - ZMB);

            FallSpeed = Average_Fallspeed (ZM, ParticleBinIndex);

            if (TM >= TS)
                ZS = ZM;
            else
                ZS = ZM - (TS - TM)*FallSpeed;

            if (ZS >= ZBF)
                RS = RF;
            else
                RS = RI + (ZS - ZMin)*TerpR;

            TG = TM + (ZM - ZGZ)/FallSpeed;


            Alt1Coeff1 = (ZM - ZI)/(sqrt (TM) - sqrt (TI));

            Alt1Coeff0 = ZI - sqrt (TI)*Alt1Coeff1;

            Alt2Coeff1 = -FallSpeed;

            Alt2Coeff0 = ZM + FallSpeed*TM;

            SqrtCutoffTime = TM;


            /*  Add 5 second tweek so Integrate never evaluates  */
            /*  Wind functions at a negative time.               */

            XG = Integrate (XWindFunctionOfTime, TI + 5.0, TG, 0.001, 0.001);
            YG = Integrate (YWindFunctionOfTime, TI + 5.0, TG, 0.001, 0.001);


            #ifdef UseConstantK

                RG = sqrt (Sqr (RS)/4.0 + Two_Kr*(Max (0.0, TG - TS)));

            #else

                if (ZS <= ZGZ)
                {
                    RG = RS/2.0;
                }
                else
                {
                    SigPow = pow (RS/2.0,   2.0/3.0);
                    SHPow  = pow (ZS - ZGZ, 2.0/3.0);

                    C = 3.83155*FallSpeed*(1000.0 - SigPow);

                    if (SHPow <= C)
                        RG = pow ((SigPow + 0.26099*SHPow/FallSpeed), 3.0/2.0);
                    else
                        RG = sqrt (7.8297e5*SHPow/FallSpeed +
                                   3.0e6*SigPow - 2.0e9);

                };

            #endif


            if (i == NCD)
            {
                /*   Except under unusual wind conditions,          */
                /*   the top parcel of the smallest particle size   */
                /*   bin will land the greatest distance from the   */
                /*   burst location.                                */

                FarX = XG;
                FarY = YG;
                FarR = RG;

            };


            PRIndex++;

            PR [PRIndex].TG = TG;
            PR [PRIndex].XG = XG;
            PR [PRIndex].YG = YG;
            PR [PRIndex].RG = RG;

            PR [PRIndex].Bin      = ParticleBinIndex;
            PR [PRIndex].Division = i;

            #ifdef Debug

                printf ("\n");

                if (i == 0)
                    printf ("Bottom of Parcel 1 \n");
                else
                    if (i == NCD)
                        printf ("Top of Parcel %1i \n", i);
                    else
                        printf ("Top of Parcel %1i / Bottom of Parcel %1i \n",
                                 i, i+1);

                printf ("\n");
                printf ("Max Altitude ZM        =  %11.2lf m."
                        " at TM = %11.2lf s. \n", ZM, TM);
                printf ("Stabilized Altitude    =  %11.2lf m. \n", ZS);
                printf ("Stabilized Radius RS   =  %11.2lf m."
                        "   Ground Sigma = %11.2lf m. \n", RS, RG);
                printf ("Time of Ground Impact  =  %11.2lf s. \n", TG);
                printf ("Ground Impact Position = %11.2lf m.,"
                        "        %11.2lf m. \n", XG, YG);

            #endif

        };


        ParticleBinIndex += NPS;

        UpdateRunProgressIndicator
            (1, 100.0*Min (1.0, 1.0*ParticleBinIndex/Total_NPS));

    }
    while (ParticleBinIndex < Total_NPS);


    NumParcelRecords = PRIndex;

};  /*  void Compute_Ground_Parameters_For_Each_Parcel  */








void CalculateGridTransformation (void)

/*  Compute transformation coefficients between grid point indices and
/*  X-Y downwind/acrosswind coordinates.
/*
/*  The point represented by the grid point Grid [i, j] can be
/*  calculated using...
/*
/*     Xdownwind   = XCoeff0 + XCoeff1 * i;
/*     Yacrosswind = YCoeff0 + YCoeff1 * j;
/*
/*  ...for 0 <= I < NumXGridPoints and 0 <= J < NumYGridPoints.
/*
/**/

{
    XCoeff0 = XGridOrigin;
    YCoeff0 = YGridOrigin;

    XCoeff1 = XGridSpacing;
    YCoeff1 = YGridSpacing;

    #ifdef Debug

        printf ("\n");
        printf ("\n");
        printf ("X-Y / Grid Indices Transformation Coefficients = ...\n");
        printf ("\n");
        printf ("   XCoeff0 = %11.4lf m     XCoeff1 = %11.4lf m \n",
                 XCoeff0, XCoeff1);
        printf ("   YCoeff0 = %11.4lf m     YCoeff1 = %11.4lf m \n",
                 YCoeff0, YCoeff1);
        printf ("\n");
        printf ("...where...\n");
        printf ("\n");
        printf ("   X = XCoeff0 + XCoeff1 * i \n");
        printf ("   Y = YCoeff0 + YCoeff1 * j \n");
        printf ("\n");

    #endif

};  /*  void CalculateGridTransformation  */








void ComputeGridTimes (void)

/*  Sets the time field of each of the NumberComputedGrids grid records in
/*  the list whose head is pointed to by the global variable GridListHead.
/*
/*  The times are set such that the time of the first grid is equal to the
/*  time of landing of the first parcel, and the time of the last grid is
/*  equal to the time of landing of the last parcel.  The other grid times
/*  are set such that an equal amount of fallout lands between the times of
/*  each grid.
/*
/*  If there is any deposition occuring at H+1 hour, the time of one of the
/*  grids is set to 1 hour, so that any hazard indicators which rely on the
/*  deposition at one hour can be exactly computed.
/*
/**/

{
    int I,
        N,
        PRIndex,
        DepoIndex,
        NCG;

    GridPointer GridP;

    int ExactlyOneHour;

    double T,
           Depo,
           Rate,
           TMin,
           TMax,
           TotalDepo,
           TargetDepo;

    double (*TimeP)        [];
    double (*DepoRateIncP) [];


    TimeP        = malloc (sizeof (double)*753);
    DepoRateIncP = malloc (sizeof (double)*753);

    N         = 0;
    NCG       = NumberComputedGrids;
    TotalDepo = 0.0;

    for (PRIndex = 1; PRIndex <= NumParcelRecords; PRIndex++)
    {
        if (PR [PRIndex].Division != 0)
        {
            TotalDepo += FP [PR [PRIndex].Bin];

            N++;

            (*TimeP) [N]        = Min (PR [PRIndex].TG, PR [PRIndex+1].TG);
            (*DepoRateIncP) [N] = FP [PR [PRIndex].Bin] /
                                  fabs (PR [PRIndex].TG - PR [PRIndex+1].TG);

            N++;

            (*TimeP) [N]        = Max (PR [PRIndex].TG, PR [PRIndex+1].TG);
            (*DepoRateIncP) [N] = -(*DepoRateIncP) [N-1];
        };
    };

    HeapSort (&((*TimeP) [1]),        N, 1);
    HeapSort (&((*DepoRateIncP) [1]), N, 3);

    TMin = (*TimeP) [1];
    TMax = (*TimeP) [N];

    (*DepoRateIncP) [N] = 9.9e7;

    N++;

    (*TimeP)        [N] = 9.9e7;
    (*DepoRateIncP) [N] = 9.9e7;


    GridP = GridListHead;

    ExactlyOneHour = ((TMin < 3600.0) && (TMax > 3600.0));

    NCG--;

    if (ExactlyOneHour) NCG--;

    Depo = 0.0;
    T    = TMin;
    Rate = (*DepoRateIncP) [1];

    DepoIndex = 1;

    for (I = 0; I <= NCG; I++)
    {
        TargetDepo = (TotalDepo*I)/NCG;

        while ((Depo + Rate*((*TimeP) [DepoIndex + 1] - T)) < TargetDepo)
        {
            DepoIndex++;

            Depo = Depo + Rate*((*TimeP) [DepoIndex] - T);
            Rate = Rate + (*DepoRateIncP) [DepoIndex];
            T    = (*TimeP) [DepoIndex];
        };

        T    = T + (TargetDepo - Depo)/Rate;
        Depo = TargetDepo;

       if (ExactlyOneHour && (T > 3600.0))
       {
           ExactlyOneHour = 0;
           GridP->Time    = 3600.0;
           GridP          = GridP->Next;
       };

       GridP->Time = T;

       GridP = GridP->Next;

    };


    #ifdef Debug

          printf ("\n");
          printf ("\n");
          printf ("Fallout Time Statistics \n");
          printf ("\n");
          printf ("\n");
          printf ("Time of First Fallout = %8.0lf s \n", TMin);
          printf ("Time of Last Fallout  = %8.0lf s \n", TMax);
          printf ("\n");
          printf ("Grids of the radiation footprint will be calculated at \n");
          printf ("the following times... \n");
          printf ("\n");
          printf ("Grid Number     H + Time (s) \n");

          GridP = GridListHead;
          I     = 0;

          while (GridP != NULL)
          {
              printf ("%6i      %12.0lf \n", ++I, GridP->Time);

              GridP = GridP->Next;
          };

    #endif


    free (TimeP);
    free (DepoRateIncP);

};  /*  void ComputeGridTimes  */








void AddToGrids (void)

{
    #define Sqr(X) (pow((X),2.0))

    const double Tiny = 0.001;

    int IX,
        IY,
        IR,
        Step,
        PRIndex,
        NextPRIndex,
        IndicatorSteps;

    long int MinIX,
             MinIY,
             MaxIX,
             MaxIY,
             IExtX1,
             IExtY1,
             IExtX2,
             IExtY2;

    double r,
           C,
           Q,
           X,
           Y,
           Xp,
           Yp,
           Xb,
           Yb,
           Sb,
           Xt,
           Yt,
           St,
           Temp,
           Tend,
           Tbegin,
           CosA,
           SinA,
           ExtX1,
           ExtY1,
           ExtX2,
           ExtY2,
           XpRot,
           YpRot,
           GaussQ,
           GaussX,
           GaussY,
           DeltaX,
           DeltaY,
           Sig2Parl,
           Sig2Perp,
           InsideExp,
           CloudSizeB,
           CloudSizeT,
           GaussCoeffX,
           GaussCoeffY;

    GridPointer GridP;

    GridArrayPtr Gp;


    #ifdef Debug

        printf ("\n");
        printf ("\n");
        printf ("\n");
        printf ("COMPUTATIONAL  PHASE  TWO \n");
        printf ("\n");
        printf ("Add each parcel's contribution to the appropriate grids. \n");
        printf ("\n");
        printf ("\n");

    #endif

    Gp = malloc (sizeof (GridArray));

    IndicatorSteps = NCD*Round ((Total_NPS + 0.1)/NPS);
    Step           = 0;

    UpdateRunProgressIndicator (2, 0.0);

    PRIndex = 0;

    while (PRIndex < NumParcelRecords)
    {

        PRIndex++;

        NextPRIndex = PRIndex + 1;

        if (PR [PRIndex].Division != 0)
        {
            Q = FP [PR [PRIndex].Bin];

            Tend = PR [PRIndex].TG;

            Xt = PR [PRIndex].XG;
            Yt = PR [PRIndex].YG;
            St = PR [PRIndex].RG;

            Tbegin = PR [NextPRIndex].TG;

            Xb = PR [NextPRIndex].XG;
            Yb = PR [NextPRIndex].YG;
            Sb = PR [NextPRIndex].RG;

            if (Tend < Tbegin)
            {
                Temp   = Tbegin;
                Tbegin = Tend;
                Tend   = Temp;
            };

            DeltaX = Xt - Xb;
            DeltaY = Yt - Yb;

            if (fabs (DeltaX ) < Tiny) DeltaX = Tiny;
                /*  Avoid /0 for zero wind speeds.  */

            r = sqrt (DeltaX*DeltaX + DeltaY*DeltaY);

            Xp = (Xt + Xb)/2.0;
            Yp = (Yt + Yb)/2.0;

            Sig2Parl = pow (St + Sb + r, 2.0)/4.0;

            Sig2Perp = St*Sb;

            CosA = fabs (DeltaX) / r;

            SinA = DeltaX * DeltaY / (r * fabs (DeltaX));

            XpRot = Xp*CosA + Yp*SinA;
            YpRot = Yp*CosA - Xp*SinA;


            if (R_Search > 3.5)
            {

                /*  Do not truncate when adding Gaussian parcels.  */

                MaxIX = NumXGridPoints - 1;
                MaxIY = NumYGridPoints - 1;
                MinIX = 0;
                MinIY = 0;
            }
            else
            {
                /*  Calculate set of grid points which are  */
                /*  significantly affected by this parcel.  */

                CloudSizeT = R_Search*St;
                CloudSizeB = R_Search*Sb;

                ExtX1 = Max (Xt + CloudSizeT, Xb + CloudSizeB);
                ExtY1 = Max (Yt + CloudSizeT, Yb + CloudSizeB);
                ExtX2 = Min (Xt - CloudSizeT, Xb - CloudSizeB);
                ExtY2 = Min (Yt - CloudSizeT, Yb - CloudSizeB);

                IExtX1 = RoundIn ((ExtX1 - XCoeff0)/XCoeff1);
                IExtY1 = RoundIn ((ExtY1 - YCoeff0)/YCoeff1);
                IExtX2 = RoundIn ((ExtX2 - XCoeff0)/XCoeff1);
                IExtY2 = RoundIn ((ExtY2 - YCoeff0)/YCoeff1);

                MaxIX = Min (NumXGridPoints - 1, 1 + Max (IExtX1, IExtX2));
                MaxIY = Min (NumYGridPoints - 1, 1 + Max (IExtY1, IExtY2));

                MinIX = Max (0, Min (IExtX1, IExtX2));
                MinIY = Max (0, Min (IExtY1, IExtY2));
            };


            GaussQ = Q / (TwoPi * sqrt (Sig2Parl*Sig2Perp));

            GaussCoeffX = -0.5/Sig2Parl;
            GaussCoeffY = -0.5/Sig2Perp;


            /*  Save parcel information for later use in     */
            /*  calculating integrated flight path dosages.  */

            PR [PRIndex].CosA        = CosA;
            PR [PRIndex].SinA        = SinA;
            PR [PRIndex].XpRot       = XpRot;
            PR [PRIndex].YpRot       = YpRot;
            PR [PRIndex].GaussQ      = GaussQ;
            PR [PRIndex].GaussCoeffX = GaussCoeffX;
            PR [PRIndex].GaussCoeffY = GaussCoeffY;


            /*  Find increment for section of grid 'close'  */
            /*  to ground impact point.                     */

            for (IX = (int) MinIX; IX <= MaxIX; IX++)
                for (IY = (int) MinIY; IY <= MaxIY; IY++)
                {
                    X = XCoeff1*IX + XCoeff0;
                    Y = YCoeff1*IY + YCoeff0;

                    GaussX = X*CosA + Y*SinA;
                    GaussY = Y*CosA - X*SinA;

                    InsideExp = GaussCoeffX*Sqr (GaussX - XpRot) +
                                GaussCoeffY*Sqr (GaussY - YpRot);

                    if (InsideExp > -50.0)
                        (*Gp) [IX][IY] = GaussQ * exp (InsideExp);
                    else
                        (*Gp) [IX][IY] = 0.0;

                };


            /*  Find increment for specific receptors.  */

            for (IR = 0; IR < NumReceptors; IR++)
            {
                X = ReceptorX [IR];
                Y = ReceptorY [IR];

                GaussX = X*CosA + Y*SinA;
                GaussY = Y*CosA - X*SinA;

                InsideExp = GaussCoeffX*Sqr (GaussX - XpRot) +
                            GaussCoeffY*Sqr (GaussY - YpRot);

                if (InsideExp > -50.0)
                    (*Gp) [IR][NumYGridPoints] = GaussQ * exp (InsideExp);
                else
                    (*Gp) [IR][NumYGridPoints] = 0.0;

            };


            /*  Add increments to running total grids.  */

            GridP = GridListHead;

            while (GridP)
            {
                if (GridP->Time > Tbegin)
                    if (GridP->Time >= Tend)
                    {
                        for (IX = (int) MinIX; IX <= MaxIX; IX++)
                            for (IY = (int) MinIY; IY <= MaxIY; IY++)
                                GridP->Grid [IX][IY] += (*Gp) [IX][IY];

                        for (IR = 0; IR < NumReceptors; IR++)
                            GridP->Grid [IR][NumYGridPoints] +=
                                (*Gp) [IR][NumYGridPoints];
                    }
                    else
                    {
                        C = (GridP->Time - Tbegin)/(Tend - Tbegin);

                        for (IX = (int) MinIX; IX <= MaxIX; IX++)
                            for (IY = (int) MinIY; IY <= MaxIY; IY++)
                                GridP->Grid [IX][IY] += C*(*Gp) [IX][IY];

                        for (IR = 0; IR < NumReceptors; IR++)
                            GridP->Grid [IR][NumYGridPoints] +=
                                C*(*Gp) [IR][NumYGridPoints];
                    };

                GridP = GridP->Next;

            };


            #ifdef Debug

                printf ("\n");
                printf ("Adding Parcel %2i, Bin %2i \n",
                         PR [PRIndex].Division, PR [PRIndex].Bin + 1);

                if ((MinIX > MaxIX) || (MinIY > MaxIY))
                    printf ("No significant grid points. \n");
                else
                {
                    printf ("Rectangle from ");
                    printf ("(%i, ",   MinIX + 1);
                    printf ("%i) to ", MinIY + 1);
                    printf ("(%i, ",   MaxIX + 1);
                    printf ("%i)",     MaxIY + 1);
                    printf ("\n");

                };
            #endif


            Step++;

            UpdateRunProgressIndicator
                (2, 100.0*Min (1.0, 1.0*Step/IndicatorSteps));

        }
        else
        {
            /*  Don't leave these undefined.  */

            PR [PRIndex].CosA        = 0.0;
            PR [PRIndex].SinA        = 0.0;
            PR [PRIndex].XpRot       = 0.0;
            PR [PRIndex].YpRot       = 0.0;
            PR [PRIndex].GaussQ      = 0.0;
            PR [PRIndex].GaussCoeffX = 0.0;
            PR [PRIndex].GaussCoeffY = 0.0;

        };

    };


    free (Gp);

};  /*  void AddToGrids  */








void PerformComputations ()

/*  Runs the Airrad computational engine on the Airrad case last loaded by
/*  a call to the function ReadCaseVariables.  The results of the run are
/*  stored in a global variable, and may be accessed by using the
/*  ComputeHazard function.  The results may also be stored and retrieved
/*  to/from disk by using the WriteComputedGrid and ReadComputedGrid
/*  functions.
/*
/*  On return, the global variable GridListHead points to a linked list of
/*  grids which contain the results of the computations.  On call to
/*  PerformCalculations, GridListHead must be either nil or point to a nil
/*  terminated linked list of GridRecords.
/*
/*  The results for specific location receptor i are stored in Grid [i, np1],
/*  where np1 = NumYGridPoints + 1.
/*
/**/

{
    #ifdef Debug

      int I;

      char BinString [4][8] = {"", "other ", "third ", "fourth "};


      printf ("\n\n --- Beginning computations --- \n\n");
      printf ("\n");
      printf ("\n");
      printf ("                  *****   AIRRAD  Runtime  Listing   *****\n");
      printf ("\n");
      printf ("\n");
      printf ("Passed Run Description Variables\n");
      printf ("\n");
      printf ("Total Yield (W)                 = %11.4lf kt \n", W);
      printf ("Fission Yield (FW)              = %11.4lf kt \n", FW);
      printf ("Fission Type (FissionType)      = %11s \n",
               FissionTypeName [FissionType]);
      printf ("Height of Burst (HOB)           = %11.4lf m \n", HOB);
      printf ("Altitude of Ground Zero (ZGZ)   = %11.4lf m \n", ZGZ);
      printf ("Ground Zero X Location (BurstX) =      0.0000 m \n");
      printf ("Ground Zero Y Location (BurstY) =      0.0000 m \n");
      printf ("\n");
      printf ("\n");
      printf ("Number of Wind Layers = %3i \n", NumberOfLayers);
      printf ("\n");
      printf ("Layer    Height (m)    Speed (m/s)    Direction (deg e of n)\n");
      printf ("\n");

      for (I=0; I < NumberOfLayers; I++)
          printf ("%3i       %7.0lf        %6.2lf         %6.2lf \n",
                   I + 1,
                   WindMeasurementHeight [I],
                   WindSpeed [I],
                   WindDirection [I]);

      printf ("\n");
      printf ("\n");
      printf ("Number of X Grid Points     = %6i \n", NumXGridPoints);
      printf ("Number of Y Grid Points     = %6i \n", NumYGridPoints);
      printf ("Grid Spacing in X Direction = %11.4lf \n", XGridSpacing);
      printf ("Grid Spacing in Y Direction = %11.4lf \n", YGridSpacing);
      printf ("Grid Origin               X = %11.4lf \n", XGridOrigin);
      printf ("Grid Origin               Y = %11.4lf \n", YGridOrigin);
      printf ("\n");

      printf ("Negligible Radius = ");

      if (R_Search > 3.5)
          printf ("Infinite \n");
      else
          printf ("%11.4lf sigmas \n", R_Search);

      printf ("\n");
      printf ("Divide cloud into %i wafer(s) for tracking. \n", NCD);
      printf ("\n");
      printf ("Use every %sparticle size bin. \n", BinString [NPS-1]);

      if (NumReceptors > 0)
      {

          printf ("\n");
          printf ("\n");
          printf ("          Specific Receptors Locations \n");
          printf ("\n");
          printf ("  Receptor Name        X Location    Y Location \n");
          printf ("\n");

          for (I=0; I < NumReceptors; I++)
              printf ("   Receptor %2i        %8.0lf m    %8.0lf m \n",
                       I + 1, ReceptorX [I], ReceptorY [I]);

      };

      printf ("\n");
      printf ("\n");

    #endif


    GridListHead = ZeroGrids (GridListHead, NumberComputedGrids,
                              NumXGridPoints, NumYGridPoints, NumReceptors);

    ComputeBurstConstants ();

    ProcessWindData ();

    Compute_Activity_Fraction_Array ();

    Compute_Ground_Parameters_For_Each_Parcel ();

    CalculateGridTransformation ();

    ComputeGridTimes ();

    AddToGrids ();

};  /*  void PerformComputations  */








void PutInt (FILE* f, int X)
{

    fwrite (&X, sizeof (int), 1, f);

};  /*  void PutInt  */








void PutFloat (FILE* f, float X)
{

    fwrite (&X, sizeof (float), 1, f);

};  /*  void PutFloat  */








void PutDouble (FILE* f, double X)
{

    fwrite (&X, sizeof (double), 1, f);

};  /*  void PutDouble  */








void WriteGrid (FILE* f, GridPointer GP)

{

    int IX, IY, IR;


    PutInt (f, GP->NumXGridPoints);
    PutInt (f, GP->NumYGridPoints);
    PutInt (f, GP->NumReceptors);

    PutDouble (f, GP->Time);
    PutDouble (f, GP->Code);

    for (IX = 0; IX < GP->NumXGridPoints; IX++)
        for (IY = 0; IY < GP->NumYGridPoints; IY++)
            PutFloat (f, GP->Grid [IX][IY]);

    for (IR = 0; IR < GP->NumReceptors; IR++)
        PutFloat (f, GP->Grid [IR][GP->NumYGridPoints]);

};  /*  void WriteGrid  */








void WriteComputedGrids (char* FileName)

/*  Writes the results of the last computation to file FileName.
/*  These results can be retrieved by a call to function
/*  ReadComputedGrids.
/*
/**/

{
    FILE* f;

    GridPointer P;


    f = fopen (FileName, "wb");

    if (f == NULL) FailExit ("Could not open computed grids file for writing.");

    PutInt (f, NumberComputedGrids);
    PutInt (f, GridListHead->NumXGridPoints);
    PutInt (f, GridListHead->NumYGridPoints);
    PutInt (f, GridListHead->NumReceptors);

    PutDouble (f, GridRotationAngle);

    P = GridListHead;

    while (P)
    {
        WriteGrid (f, P);
        P = P->Next;
    };

    fclose (f);

};  /*  void WriteComputedGrids  */








