
#include "aradengn.h"








                /*  Radiation Height Attenuation Curve Coeffs.  */
                /*  Z, A, B, C, D for spline interpolation.     */

double Atten_Z [25] =
    {  0, 25, 50, 75,100,125,150,175,200,225,250,275,300,325,350,375,400,425,
     450,475,500,600,700,800,900};

    /*  Z's, heights in meters  */


double Atten_A [25] =
    {0.1000E+01,0.6101E+00,0.4280E+00,0.3097E+00,0.2378E+00,0.1857E+00,0.1475E+00,
     0.1167E+00,0.9514E-01,0.7764E-01,0.6301E-01,0.5314E-01,0.4403E-01,0.3649E-01,
     0.3117E-01,0.2624E-01,0.2195E-01,0.1888E-01,0.1615E-01,0.1366E-01,0.1145E-01,
     0.6308E-02,0.3479E-02,0.1941E-02,0.1119E-02};

double Atten_B [25] =
    {-0.1975E-01,-0.9234E-02,-0.5202E-02,-0.3312E-02,-0.2269E-02,-0.1683E-02,
     -0.1347E-02,-0.9924E-03,-0.7391E-03,-0.6470E-03,-0.4346E-03,-0.3720E-03,
     -0.3484E-03,-0.2298E-03,-0.2010E-03,-0.1913E-03,-0.1392E-03,-0.1116E-03,
     -0.1039E-03,-0.9426E-04,-0.6304E-04,-0.3658E-04,-0.1842E-04,-0.9776E-05,
     -0.5840E-05};

double Atten_C [25] =
    {0.7802E-04,0.7276E-04,-0.1962E-04,0.1093E-04,-0.1682E-05,0.5100E-05,0.6515E-08,
     0.5356E-05,0.1000E-05,-0.1070E-05,0.2269E-05,-0.3725E-07,0.8856E-06,0.9081E-06,
     0.2593E-07,0.3201E-06, 0.8656E-06,-0.3478E-07,0.1479E-06,0.2466E-06,0.8293E-07,
     0.6716E-07,0.4925E-08, 0.7177E-08,0.5977E-08};

double Atten_D [25] =
    {0.3528E-05, 0.2106E-06,0.1531E-05,0.2647E-06, 0.3574E-06,0.4277E-07,
     0.1892E-06,-0.7752E-08,0.2247E-07,0.1418E-06,-0.2711E-07,0.1358E-07,
     0.3966E-07,-0.8855E-08,0.4467E-08,0.1928E-07,-0.8385E-08,0.5022E-08,
     0.1214E-08,-0.5071E-09,0.3292E-09,0.1577E-09, 0.2553E-09,0.8336E-10,
     0.5957E-10};




void GetInt (FILE* f, int* X)
{
fopen
    fread (X, sizeof (int), 1, f);

};  /*  void GetInt  */








void GetFloat (FILE* f, float* X)
{

    fread (X, sizeof (float), 1, f);

};  /*  void GetFloat  */








void GetDouble (FILE* f, double* X)
{

    fread (X, sizeof (double), 1, f);

};  /*  void GetDouble  */








void ReadGrid (FILE* f, GridPointer GP)

{

    int IX, IY, IR;


    GetInt (f, &(GP->NumXGridPoints));
    GetInt (f, &(GP->NumYGridPoints));
    GetInt (f, &(GP->NumReceptors));

    GetDouble (f, &(GP->Time));
    GetDouble (f, &(GP->Code));

    for (IX = 0; IX < GP->NumXGridPoints; IX++)
        for (IY = 0; IY < GP->NumYGridPoints; IY++)
            GetFloat (f, &(GP->Grid [IX][IY]));

    for (IR = 0; IR < GP->NumReceptors; IR++)
       GetFloat (f, &(GP->Grid [IR][GP->NumYGridPoints]));

};  /*  void ReadGrid  */








void ReadComputedGrids (char* FileName)

/*  Reads the results of an Airrad computation from file FileName.
/*  These results should have been written with a call to function
/*  WriteComputedGrids.
/*
/**/

{
    FILE* f;

    GridPointer P;

    int NG, NX, NY, NR;


    f = fopen (FileName, "rb");

    if (f == NULL) FailExit ("Could not open computed grids file for reading.");

    GetInt (f, &NG);
    GetInt (f, &NX);
    GetInt (f, &NY);
    GetInt (f, &NR);

    GetDouble (f, &GridRotationAngle);

    GridListHead = ZeroGrids (GridListHead, NG, NX, NY, NR);

    P = GridListHead;

    while (P)
    {
        ReadGrid (f, P);
        P = P->Next;
    };

    fclose (f);

};  /*  void ReadComputedGrids  */








double Attenuation (double Height)

/*  Attenuation uses spline interpolation to determine the attenuation
/*  factor for gamma radiation (typical fallout energies) due to height
/*  above an infinite plane.
/*
/*  If I(0) is the radiation intensity at a point just above a plane
/*  covered with a uniform layer of activity, then the radiation at a point
/*  z meters above the plane is I(z) = I(0) * Attenuation (z).
/*
/*  Attenuation (z) is calculated up to 1000 meters.  Above that, it is
/*  assumed to be zero.  In reality, it is less than 1/1000.
/*
/**/

{
    double H_Diff;

    int Level;


    Height = fabs (Height);

    if (Height >= 1000.0)
    {
        return (0.0);
    }
    else
    {
        if (Height < 500.0)
            Level = RoundIn (Height/ 25.0) + 0;
        else
            Level = RoundIn (Height/100.0) + 15;

        H_Diff = Height - Atten_Z [Level];

        return (((  Atten_D [Level] * H_Diff
                  + Atten_C [Level] * H_Diff)
                  + Atten_B [Level] * H_Diff)
                  + Atten_A [Level]);

    };

};  /*  double Attenuation  */








void GetIntegrateCoeffs (double Tr1, double t1, double t2, double Tr2,
                         double* C1, double* C2)

/*  Calculates two coeffs C1 and C2, such that the total integrated exposure
/*  at a point from time t1 to time t2, in roentgens, is C1*r1 + C2*r2, where
/*  r1 is the exposure rate in roentgens/hr at time Tr1, r2 is the exposure
/*  rate at time Tr2, and both r1 and r2 are not corrected for radiation decay.
/*
/*  It is assumed that the non-decay corrected radiation intensity varies
/*  linearly from r1 to r2 over the time period Tr1 to Tr2 (ie, activity is
/*  deposited at a constant rate), and that the radiation decay factor is
/*  t^-1.26, where t is the time in hours since the burst.
/*
/*  All passed times are in seconds since the burst time.  They inequality
/*  Tr1 <= t1 < t2 <= Tr2 should hold.
/*
/**/

{
    double T26, T74, T2674;


    if (t1 >= t2)
    {
        (*C1) = 0.0;
        (*C2) = 0.0;
    }
    else
    {
        /*  Convert to hours  */

        t1 =  Max (0.02, t1/3600.0);
        t2 =  Max (0.02, t2/3600.0);

        Tr1 = Tr1/3600.0;
        Tr2 = Tr2/3600.0;

        T26 = (pow (t1, -0.26) - pow (t2, -0.26))/0.26;
        T74 = (pow (t2,  0.74) - pow (t1,  0.74))/0.74;

        T2674 = (Tr1*T26 - T74)/(Tr2 - Tr1);

        (*C1) = T26 + T2674;
        (*C2) = -T2674;
    };

};  /*  void GetIntegrateCoeffs  */








void DecayGrid (GridArray OldGrid, double Time, int NX, int NY, int NR,
                double HeightAttenuationFactor, GridArray NewGrid)

/*  Accepts a grid OldGrid of radiation intensities, with
/*  grid points IX = 0..NX-1 and IY = 0..NY-1 defined, and with NR
/*  specific receptors stored at locations [1, NY] to [NR, NY],
/*  and the time of this grid Time, in seconds w.r.t. burst time.
/*  Returns as NewGrid the grid with the proper (Time^-1.26) radiation
/*  decay factor applied.
/*
/**/

{
    int IX,
        IY,
        IR;

    double Decay;


    Decay = HeightAttenuationFactor * pow (Max (0.02, Time/3600.0), -1.26);

    for (IX = 0; IX < NX; IX++)
        for (IY = 0; IY < NY; IY++)
            NewGrid [IX][IY] = Decay*OldGrid [IX][IY];

    for (IR = 0; IR < NR; IR++)
        NewGrid [IR][IY] = Decay*OldGrid [IR][IY];

};  /*  void DecayGrid  */








void ComputeHazard (int HazardCode, double T1, double T2, double Height,
                    int RadUnits, GridArray Grid, double* GridAngle)

/*  Computes the grid hazard whose code number is HazardCode (see list below)
/*  from the list of radiation intensity grids whose head is passed as
/*  GridListHead.  The computed grid is returned via Grid.
/*
/*  RadUnits is a code word that controls the units of the returned grids.
/*  If RadUnits = 0, then all units are returned in Roentgens, Roentgens/Hr,
/*  and Seconds.  If RadUnits = 1, then all unis are returned in Rads,
/*  Rads/Hr, and Seconds.
/*
/*  T1 and T2 are two times, in seconds, which are parameters to some
/*  the hazard indicators.  See list below.
/*
/*  Hazard
/*   Code           Description
/*
/*     1       Normalized exposure rate at time H + T1 seconds. (Roentgens/Hr)
/*             T1 will usually be 3600 seconds (1 hour).
/*
/*     7       Normalized dosage between H + T1 and H + T2 seconds. (Roentgens)
/*
/*     2       Exposure Rate at time H + T1 seconds, accounting for time of
/*             arrival of fallout. (Roentgens/Hr)
/*
/*     3       Integrated Exposure from time H + T1 seconds to H + T2 seconds,
/*             accounting for time of arrival of fallout. (Roentgens)
/*
/*     4       Integrated Exposure from time H + T1 seconds to infinity,
/*             accounting for time of arrival of fallout. (Roentgens)
/*
/*     5       Time of first fallout. (Seconds w.r.t. Burst Time)
/*
/*     6       Time of last fallout. (Seconds w.r.t. Burst Time)
/*
/*
/*  Height is the distance above ground level, in meters, of the horizontal
/*  plane for which the grid is being calculated.
/*
/*  GridAngle is the angle that the X axis of the grid has been rotated to
/*  in order for it to be parallel with the mean wind direction.  It is
/*  returned in radians counter-clockwise.  0 means that the positive X (ie, the
/*  i axis) axis is pointed east, while Pi/2 means that the X axis is pointed
/*  true north.
/*
/*  The hazard values for the specific receptors are also returned
/*  by this function.  They are located in grid at [i][ny], where
/*  0 <= i < Number of Receptors, and ny = the number of y grid points.
/*
/**/

{
    GridPointer P,
                NextP,
                LastP,
                LowerP,
                UpperP;

    double C1,
           C2,
           Decay,
           FirstT,
           LastT,
           DeltaT,
           HeightAttenuationFactor;

    int IX,
        IY,
        IR,
        IYR,
        NumReceptors,
        NumXGridPoints,
        NumYGridPoints;

    double RadUnitsConversionFactor [2] = {1.0, 0.96};


    *GridAngle = GridRotationAngle;

    HeightAttenuationFactor = Attenuation (Height) *
                              RadUnitsConversionFactor [RadUnits];


    NumXGridPoints = GridListHead->NumXGridPoints;
    NumYGridPoints = GridListHead->NumYGridPoints;
    NumReceptors   = GridListHead->NumReceptors;

    IYR = NumYGridPoints;

    FirstT = Max (1.0, GridListHead->Time);


    LastP = GridListHead;

    while (LastP->Next) LastP = LastP->Next;

    LastT = LastP->Time;


    P = GridListHead;


    switch (HazardCode)
    {

    case 1: DecayGrid (LastP->Grid, T1, NumXGridPoints, NumYGridPoints,
                       NumReceptors, HeightAttenuationFactor, Grid);

            break;


    case 7: T1 = Max (T1, 72.0);
            T2 = Max (T1, T2);

            HeightAttenuationFactor = HeightAttenuationFactor *
               (pow (T1/3600.0, -0.26) - pow (T2/3600.0, -0.26))/0.26;

            for (IX = 0; IX < NumXGridPoints; IX++)
                for (IY = 0; IY < NumYGridPoints; IY++)
                    Grid [IX][IY] = HeightAttenuationFactor *
                                    LastP->Grid [IX][IY];

            for (IR = 0; IR < NumReceptors; IR++)
                Grid [IR][IYR] = HeightAttenuationFactor*LastP->Grid [IR][IYR];

            break;


    case 2: if (T1 <= FirstT)

                DecayGrid (GridListHead->Grid, T1, NumXGridPoints,
                           NumYGridPoints, NumReceptors,
                           HeightAttenuationFactor, Grid);

            else if (LastT <= T1)

                DecayGrid (LastP->Grid, T1, NumXGridPoints, NumYGridPoints,
                           NumReceptors, HeightAttenuationFactor, Grid);

            else
            {
                while (P->Next->Time < T1) P = P->Next;

                LowerP = P;
                UpperP = P->Next;

                Decay = HeightAttenuationFactor *
                        pow (Max (0.02, T1/3600.0), -1.26);

                DeltaT = UpperP->Time - LowerP->Time;

                C1 = Decay*(UpperP->Time - T1)/DeltaT;
                C2 = Decay*(T1 - LowerP->Time)/DeltaT;

                for (IX = 0; IX < NumXGridPoints; IX++)
                    for (IY = 0; IY < NumYGridPoints; IY++)
                        Grid [IX][IY] = C1*LowerP->Grid [IX][IY] +
                                        C2*UpperP->Grid [IX][IY];

                for (IR = 0; IR < NumReceptors; IR++)
                    Grid [IR][IYR] = C1*LowerP->Grid [IR][IYR] +
                                     C2*UpperP->Grid [IR][IYR];
            };

            break;


    case 3:
    case 4: if (HazardCode == 4) T2 = 9.9e30;

            for (IX = 0; IX < NumXGridPoints; IX++)
                for (IY = 0; IY < NumYGridPoints; IY++)
                    Grid [IX][IY] = 0.0;

            for (IR = 0; IR < NumReceptors; IR++)
                Grid [IR][IYR] = 0.0;

            while (P)
            {
                P->Code = 0.0;
                P       = P->Next;
            };


            T1 = Max (T1, FirstT);
            T2 = Max (T1, T2);


            P = GridListHead;

            while ((P->Next) && (P->Next->Time <= T1))
                P = P->Next;


            while ((P->Next) && (P->Time <= T2))
            {
                NextP = P->Next;

                GetIntegrateCoeffs (P->Time, Max (T1, P->Time),
                                    Min (T2, NextP->Time),
                                    NextP->Time, &C1, &C2);

                P->Code     = P->Code + C1;
                NextP->Code = NextP->Code + C2;

                P = NextP;

            };

            if (T2 > LastT)
                LastP->Code += (pow (Max (T1, LastT)/3600.0, -0.26) -
                                pow (T2/3600.0, -0.26)) / 0.26;


            P = GridListHead;

            while (P)
            {
                if (P->Code != 0.0)
                {
                    C1 = P->Code*HeightAttenuationFactor;

                    for (IX = 0; IX < NumXGridPoints; IX++)
                        for (IY = 0; IY < NumYGridPoints; IY++)
                            Grid [IX][IY] += C1*P->Grid [IX][IY];

                    for (IR = 0; IR < NumReceptors; IR++)
                        Grid [IR][IYR] += C1*P->Grid [IR][IYR];

                };

                P = P->Next;
            };

            break;


    case 5: for (IX = 0; IX < NumXGridPoints; IX++)
                for (IY = 0; IY < NumYGridPoints; IY++)
                    Grid [IX][IY] = 0.0;

            for (IR = 0; IR < NumReceptors; IR++)
                Grid [IR][IYR] = 0.0;

            while (P)
            {
                for (IX = 0; IX < NumXGridPoints; IX++)
                    for (IY = 0; IY < NumYGridPoints; IY++)
                         if ((Grid [IX][IY] == 0.0) &&
                             (P->Grid [IX][IY] != 0.0))
                             Grid [IX][IY] = P->Time;

                for (IR = 0; IR < NumReceptors; IR++)
                    if ((Grid [IR][IYR] == 0.0) && (P->Grid [IR][IYR] != 0.0))
                        Grid [IR][IYR] = P->Time;

                P = P->Next;
            };

            break;


    case 6: for (IX = 0; IX < NumXGridPoints; IX++)
                for (IY = 0; IY < NumYGridPoints; IY++)
                    Grid [IX][IY] = 0.0;

            for (IR = 0; IR < NumReceptors; IR++)
                Grid [IR][IYR] = 0.0;

            while (P->Next)
            {
                NextP = P->Next;

                for (IX = 0; IX < NumXGridPoints; IX++)
                    for (IY = 0; IY < NumYGridPoints; IY++)
                        if ((NextP->Grid [IX][IY] > 0.0) &&
                            (P->Grid [IX][IY] != NextP->Grid [IX][IY]))
                            Grid [IX][IY] = NextP->Time;

                for (IR = 0; IR < NumReceptors; IR++)
                    if ((NextP->Grid [IR][IYR] > 0.0) &&
                        (P->Grid [IR][IYR] != NextP->Grid [IR][IYR]))
                        Grid [IR][IYR] = NextP->Time;

                P = P->Next;

            };

            break;


     default: Grid = GridListHead->Grid;

    };  /*  switch (Code)...  */

};  /*  void ComputeHazard  */








