
Uses Local, Declerat, Strings, Engine, Graph;



Function DecodeFissionType (S: String): Integer;

Var i: Integer;

Begin

    For i := 0 to 7 do
       If (Pos (FissionTypeName [i], S) > 0)
       Then
         Begin
            DecodeFissionType := i;
            Exit;
         End;

    Writeln ('Unrecognized fission type');

    Halt (1);

End;  (*  int DecodeFissionType  *)








Procedure ReadCaseVariables (FileName: String);

Var ION: Text;

    S: String;

    Zmod: Double;

    X1, X2, Y1, Y2: Double;  (*  Coordinates of lower-left and upper-right  *)
                             (*  corners of grid, in meters alongwind and   *)
                             (*  across-wind from the burst locaton.        *)

    i: Integer;

Begin

    Assign (ION, FileName);

    {$I-}   Reset (ION);   {$I+}

    If (IOResult <> 0)
    Then
      Begin
         Writeln ('Could not open input file named "', FileName, '"');
         Halt (1);
      End;

     RD.Julian  := 0;
     RD.CaseRun := False;

    Readln (ION);  (*  Skip #Title line  *)

    Readln (ION);  (*  Read case title  *)


    Readln (ION);  (*  Skip #Burst line  *)

    Readln (ION, RD.HOB, RD.W, RD.FW, S);

    RD.FissionType := DecodeFissionType (s);


    Readln (ION);  (*  Skip #Meteorology line  *)

    Readln (ION, S);

    S := Strip (S, 'Both');

    RD.NumberOfLayers := Round (GetNumber (S, 2));

    i := Length (S);

    While (S [i] <> ' ') do Dec (i);

    RD.ZGZ := GetNumber (Copy (S, i, 255), 1);

    If (Pos ('msl', S) > 0)
       Then Zmod := -RD.ZGZ
       Else Zmod := 0.0;


    For i := 1 to RD.NumberOfLayers do
      Begin
         Readln (ION, RD.WindMeasurementHeight [i], RD.WindDirection [i], RD.WindSpeed [i]);

         RD.WindMeasurementHeight [i] := RD.WindMeasurementHeight [i] + Zmod;

      End;


    Readln (ION);  (*  Skip #Grid line  *)

    Readln (ION, RD.NumXGridPoints, RD.NumYGridPoints, X1, X2, Y1, Y2);

    RD.XGridOrigin := X1;
    RD.YGridOrigin := Y1;

    RD.XGridSpacing := (X2 - X1) / (RD.NumXGridPoints-1);
    RD.YGridSpacing := (Y2 - Y1) / (RD.NumYGridPoints-1);

    RD.BurstX := 0;
    RD.BurstY := 0;


    Readln (ION);  (*  Skip #Model line  *)

    Readln (ION, RD.R_Search, RD.NPS, RD.NCD);

    If (RD.R_Search < 0.1) Then RD.R_Search := 9.9e99;


    Readln (ION);  (*  Skip #Receptors line  *)

    Readln (ION, RD.NumReceptors);

    For i := 1 to RD.NumReceptors do
      Begin
         Str (i:1, RD.ReceptorName [i]);
         Readln (ION, RD.ReceptorX [i], RD.ReceptorY [i]);
      End;


   Close (ION);

End;  (*  Procedure ReadCaseVariables  *)










Procedure WriteGrids (GridListHead: GridPointer; FileName: String);

Var IOT: File;

     Procedure PutInteger (X: Integer);
     Begin
          BlockWrite (IOT, X, SizeOf (Integer));
     End;

     Procedure PutSingle (X: Single);
     Begin
          BlockWrite (IOT, X, SizeOf (Single));
     End;

     Procedure PutDouble (X: Double);
     Begin
          BlockWrite (IOT, X, SizeOf (Double));
     End;

     Procedure WriteGrid (GP: GridPointer);

     Var IX, IY, IR: Integer;

     Begin

         PutInteger (GP^.NumXGridPoints);
         PutInteger (GP^.NumYGridPoints);
         PutInteger (GP^.NumReceptors);

         PutDouble (GP^.Time);
         PutDouble (GP^.Code);

         For IX := 1 to GP^.NumXGridPoints do
             For IY := 1 to GP^.NumYGridPoints do
                 PutSingle (GP^.Grid [IX, IY]);

         For IR := 1 to GP^.NumReceptors do
            PutSingle (GP^.Grid [IR, GP^.NumYGridPoints + 1]);

     End;  (*  void WriteGrid  *)


Var P: GridPointer;

Begin

     Assign (IOT, FileName);

     Rewrite (IOT, 1);

     PutInteger (NumberComputedGrids);
     PutInteger (GridListHead^.NumXGridPoints);
     PutInteger (GridListHead^.NumYGridPoints);
     PutInteger (GridListHead^.NumReceptors);

     P := GridListHead;

     While (P <> Nil) do
       Begin
          WriteGrid (P);
          P := P^.Next;
       End;

     Close (IOT);

End;     (*   Procedure WriteGrids   *)










Var AbortCode    : Byte;
    GridListHead : GridPointer;



Begin

     CL.StdBdr    := Green;
     CL.StdTit    := Green;
     CL.StdTxt    := Green;
     CL.StdSpc    := LightGreen;
     CL.InfoBdr   := Green;
     CL.InfoTit   := Green;
     CL.InfoTxt   := Green;
     CL.InfoSpc   := LightGreen;
     CL.RunTimCol := Red shl 4 + Green;

     ReadCaseVariables (ParamStr (1));

     Writeln ('Beginning Calcs');

     PerformCalculations (RD, 3, 'pascal.txt', GridListHead, AbortCode);

     If (AbortCode <> 0)
     Then
       Begin
          Writeln ('ERROR - AbortCode = ', AbortCode);
          Halt (1);
       End;

     WriteGrids (GridListHead, 'pascal.grd');

     Writeln ('Ending Calcs');

End.