
Uses Local, Declerat, Engine, Hazard, Strings, Math, Graph;




Var IOT: File;
Var IOR: Text;



Procedure PutInteger (X: Integer);
Begin
     BlockWrite (IOT, X, SizeOf (Integer));
End;

Procedure PutSingle (X: Single);
Begin
     BlockWrite (IOT, X, SizeOf (Single));
End;

Procedure PutDouble (X: Double);
Begin
     BlockWrite (IOT, X, SizeOf (Double));
End;

Procedure WriteGrid (NX, NY, NR: Integer; Var Grid: GridArray);

Var IX, IY, IR: Integer;

Begin

    PutInteger (NX);
    PutInteger (NY);
    PutInteger (NR);

    PutDouble (0);
    PutDouble (0);

    For IX := 1 to NX do
        For IY := 1 to NY do
          Begin
             Writeln (IOR, IX:2, ' ', IY:2, ' ', Grid [IX, IY]:11:5);
             PutSingle (Grid [IX, IY]);
          End;

    For IR := 1 to NR do
      Begin
         Writeln (IOR, IR:2, ' ', Grid [IR, NY+1]:11:5);
         PutSingle (Grid [IR, NY + 1]);
      End;

End;  (*  void WriteGrid  *)









Procedure LoadGridFile (FileName: String; Var ErrorCode: Byte);

(*   Reads in a grid file an constructs a linked list of grid records.
(*
(*   ErrorCode = 0 - Read successful.
(*   ErrorCode = 1 - File not found or not successfully opened.
(*   ErrorCode = 2 - File found, but was not a grid file.
(*
(**)

Var ION: File of GridRecord;

    IX,
    IY,
    PRIndex: Integer;

    P: GridPointer;

    G: GridRecord;

    V: Single;





Function GetValue: Single;

Begin

     GetValue := G.Grid [IX, IY];

     Inc (IX);

     If (IX > MaxXGridSize)
     Then
       Begin

          IX := 1;

          Inc (IY);

          If (IY > MaxYGridSizePlus1)
          Then
            Begin

               IY := 1;

               Read (ION, G);

            End;

        End;

End;      (*   Single Function GetValue   *)





Function FileCheck (FileName: String; FileType: IDString): Byte;

(*   Checks on file FileName.
(*
(*   Returns 1 if the file does not exist,
(*   2 if the file exists but is not of the file type FileType,
(*   and 0 otherwise.
(*
(**)

Var ION: File of IDString;

    ID_Line: IDString;

Begin

     Assign (ION, FileName);

     {$I-} Reset (ION); {$I+}

     If (IOResult <> 0)
     Then
       Begin
          FileCheck := 1;
          Exit;
       End;

     {$I-} Read (ION, ID_Line); {$I+}

     If (IOResult <> 0)
     Then
       Begin
          FileCheck := 1;
          Exit;
       End

     Else If (ID_Line = FileType)
             Then FileCheck := 0
             Else FileCheck := 2;

     Close (ION);

End;      (*   Byte Function FileCheck   *)










Begin

     (*   Get rid of old grids.   *)

     While (GridHead <> Nil) do
       Begin

          Dispose (GridHead);

          GridHead := GridHead^.Next;

       End;

     GridHead := Nil;


     ErrorCode := FileCheck (FileName, GridIDString);

     If (ErrorCode <> 0) Then Exit;


     Assign (ION, FileName);

     Reset (ION);


     (*   Read Parcel Record Array - must unpack grid records.   *)

     IX := 1;
     IY := 1;

     Read (ION, G);

     NumParcelRecords := G.NumReceptors;

     For PRIndex := 1 to NumParcelRecords do
       Begin

          PR [PRIndex].GaussQ      := GetValue;
          PR [PRIndex].GaussCoeffX := GetValue;
          PR [PRIndex].GaussCoeffY := GetValue;
          PR [PRIndex].XpRot       := GetValue;
          PR [PRIndex].YpRot       := GetValue;
          PR [PRIndex].CosA        := GetValue;
          PR [PRIndex].SinA        := GetValue;
          PR [PRIndex].TG          := GetValue;
          V                        := GetValue;

          PR [PRIndex].Division := Round (V / 1000);

          PR [PRIndex].Bin := Round (V - PR[PRIndex].Division*1000);

       End;



     While Not Eof (ION) do
       Begin

          If (GridHead = Nil)

          Then
            Begin
               New (GridHead);
               P := GridHead;
            End

          Else
            Begin
               New (P^.Next);
               P := P^.Next;
            End;

         Read (ION, P^);

         P^.Next := Nil;

       End;

     Close (ION);

End;      (*   Procedure LoadGridFile   *)







Function DecodeFissionType (S: String): Integer;

Var i: Integer;

Begin

    For i := 0 to 7 do
       If (Pos (FissionTypeName [i], S) > 0)
       Then
         Begin
            DecodeFissionType := i;
            Exit;
         End;

    Writeln ('Unrecognized fission type');

    Halt (1);

End;  (*  int DecodeFissionType  *)








Procedure ReadCaseVariables (FileName: String);

Var ION: Text;

    S: String;

    Zmod: Double;

    X1, X2, Y1, Y2: Double;  (*  Coordinates of lower-left and upper-right  *)
                             (*  corners of grid, in meters alongwind and   *)
                             (*  across-wind from the burst locaton.        *)

    i: Integer;

Begin

    Assign (ION, FileName);

    {$I-}   Reset (ION);   {$I+}

    If (IOResult <> 0)
    Then
      Begin
         Writeln ('Could not open input file named "', FileName, '"');
         Halt (1);
      End;

     RD.Julian  := 0;
     RD.CaseRun := False;

    Readln (ION);  (*  Skip #Title line  *)

    Readln (ION);  (*  Read case title  *)


    Readln (ION);  (*  Skip #Burst line  *)

    Readln (ION, RD.HOB, RD.W, RD.FW, S);

    RD.FissionType := DecodeFissionType (s);


    Readln (ION);  (*  Skip #Meteorology line  *)

    Readln (ION, S);

    S := Strip (S, 'Both');

    RD.NumberOfLayers := Round (GetNumber (S, 2));

    i := Length (S);

    While (S [i] <> ' ') do Dec (i);

    RD.ZGZ := GetNumber (Copy (S, i, 255), 1);

    If (Pos ('msl', S) > 0)
       Then Zmod := -RD.ZGZ
       Else Zmod := 0.0;


    For i := 1 to RD.NumberOfLayers do
      Begin
         Readln (ION, RD.WindMeasurementHeight [i], RD.WindDirection [i], RD.WindSpeed [i]);

         RD.WindMeasurementHeight [i] := RD.WindMeasurementHeight [i] + Zmod;

      End;


    Readln (ION);  (*  Skip #Grid line  *)

    Readln (ION, RD.NumXGridPoints, RD.NumYGridPoints, X1, X2, Y1, Y2);

    RD.XGridOrigin := X1;
    RD.YGridOrigin := Y1;

    RD.XGridSpacing := (X2 - X1) / (RD.NumXGridPoints-1);
    RD.YGridSpacing := (Y2 - Y1) / (RD.NumYGridPoints-1);

    RD.BurstX := 0;
    RD.BurstY := 0;


    Readln (ION);  (*  Skip #Model line  *)

    Readln (ION, RD.R_Search, RD.NPS, RD.NCD);

    If (RD.R_Search < 0.1) Then RD.R_Search := 9.9e99;


    Readln (ION);  (*  Skip #Receptors line  *)

    Readln (ION, RD.NumReceptors);

    For i := 1 to RD.NumReceptors do
      Begin
         Str (i:1, RD.ReceptorName [i]);
         Readln (ION, RD.ReceptorX [i], RD.ReceptorY [i]);
      End;


   Close (ION);

End;  (*  Procedure ReadCaseVariables  *)










Var ION          : Text;
    HazCode      : Integer;
    T1, T2       : Double;
    Height       : Double;
    Grid         : GridArray;
    NX, NY, NR   : Integer;
    i, N         : Integer;
    ErrorCode    : Byte;
    AbortCode    : Byte;

Begin

     GridHead := Nil;

     Assign (IOR, 'report.txt');

     Rewrite (IOR);

     Assign (ION, 'j:\airrad\ansic\hazscens.dat');

     {$I-}   Reset (ION);   {$I+}

     If (IOResult <> 0)
     Then
       Begin
          Writeln ('ERROR - Could not open input file "HAZSCENS.DAT"');
          Halt (1);
       End;


     Assign  (IOT, 'hazscens.grd');
     Rewrite (IOT, 1);


     CL.StdBdr    := Green;
     CL.StdTit    := Green;
     CL.StdTxt    := Green;
     CL.StdSpc    := LightGreen;
     CL.InfoBdr   := Green;
     CL.InfoTit   := Green;
     CL.InfoTxt   := Green;
     CL.InfoSpc   := LightGreen;
     CL.RunTimCol := Red shl 4 + Green;

     CR.OutRadUnits := 1;

     ReadCaseVariables (ParamStr (1));

     Writeln ('Beginning Calcs');

     PerformCalculations (RD, 3, 'pascal.txt', GridHead, AbortCode);

     If (AbortCode <> 0)
     Then
       Begin
          Writeln ('ERROR - AbortCode = ', AbortCode);
          Halt (1);
       End;

     Writeln ('Ending Calcs');





     NX := GridHead^.NumXGridPoints;
     NY := GridHead^.NumYGridPoints;
     NR := GridHead^.NumReceptors;


     Readln (ION, N);

     For i := 1 to N do
       Begin
          Readln (ION, HazCode, T1, T2, Height);
          ComputeHazardGrid (HazCode, T1, T2, Height, GridHead, Grid);
          WriteGrid (NX, NY, NR, Grid);
       End;

     Close (ION);
     Close (IOT);
     Close (IOR);

     Writeln ('Done');

     Wait;

End.
