/*
/*  WindTest - Program to verify that the missing wind extrapolation
/*             procedures in the Airrad 8.0 computational engines are
/*             equivalent in the Pascal and ANSI C versions.
/*
/*  Author: Fred Wasmer
/*          Wasmer Consulting
/*
/*  Language: ANSI C
/*
/*  Created: July, 1994
/*
/*
/**/


#include "aradengn.h"




FILE* fin1;
FILE* fin2;
FILE* fout;








void InputProfile (FILE* f, WindArrayType H, WindArrayType S,
                   WindArrayType D, int* N)
{
    int i;


    fscanf (f, " %i", N);

    for (i=0; i < *N; i++)
        fscanf (f, " %lf %lf %lf", &(H [i]), &(S [i]), &(D [i]));

};  /*  void InputProfile  */








void TestProfile (int ProfileCount)

{

    #define  NoMatch(X,Y)  (fabs((X)-(Y)) > 0.001)

    static WindArrayType A2, U2, V2;

    int N2, i;


    fprintf (fout, "Testing profile %i. \n\n", ProfileCount);

    InputProfile (fin1, WindMeasurementHeight, WindSpeed, WindDirection, &NumberOfLayers);

    ZGZ = 100.0;

    ProcessWindData ();

    InputProfile (fin2, A2, U2, V2, &N2);

    if (NumberOfLayers != N2)
        FailExit ("Number of layers doesn't match");

    for (i = 0; i < NumberOfLayers; i++)
        if ( NoMatch (WindMeasurementAltitude [i], A2 [i]) ||
             NoMatch (Uwind [i], U2 [i]) ||
             NoMatch (Vwind [i], V2 [i])
           )
            FailExit ("Layer doesn't match");

};  /*  void TestProfile  */








int main ()
{
    int NumProfs, i;


    fin1 = fopen ("rawprofs.dat", "r");
    fin2 = fopen ("pasprofs.dat", "r");
    fout = fopen ("windtest.out", "w");

    fscanf (fin1, " %i", &NumProfs);

    for (i=1; i<=NumProfs; i++)
        TestProfile (i);

    fclose (fin1);
    fclose (fin2);
    fclose (fout);

    printf ("All profiles matched. \n");

    wait;

    return (0);

};  /*  int main  */

